@@ -76,14 +76,21 @@ static struct gotd_session_write {
 	pid_t pid;
 	const char *title;
 	struct got_repository *repo;
-	struct gotd_repo *repo_cfg;
+	char repo_name[NAME_MAX];
 	int *pack_fds;
 	int *temp_fds;
+	int content_fd;
 	struct gotd_imsgev parent_iev;
 	struct gotd_imsgev notifier_iev;
 	struct timeval request_timeout;
 	enum gotd_session_write_state state;
 	struct gotd_imsgev repo_child_iev;
+	struct got_pathlist_head notification_refs;
+	struct got_pathlist_head notification_ref_namespaces;
+	size_t num_notification_refs_needed;
+	size_t num_notification_refs_received;
+	struct got_pathlist_head *notification_refs_cur;
+	struct gotd_notification_targets notification_targets;
 } gotd_session;
 
 static struct gotd_session_client {
@@ -113,7 +120,7 @@ disconnect(struct gotd_session_client *client)
 	log_debug("uid %d: disconnecting", client->euid);
 
 	if (gotd_imsg_compose_event(&gotd_session.parent_iev,
-	    GOTD_IMSG_DISCONNECT, PROC_SESSION_WRITE, -1, NULL, 0) == -1)
+	    GOTD_IMSG_DISCONNECT, GOTD_PROC_SESSION_WRITE, -1, NULL, 0) == -1)
 		log_warn("imsg compose DISCONNECT");
 
 	imsgbuf_clear(&gotd_session.repo_child_iev.ibuf);
@@ -148,8 +155,8 @@ disconnect_on_error(struct gotd_session_client *client,
 		if (imsgbuf_init(&ibuf, client->fd) == -1) {
 			log_warn("imsgbuf_init");
 		} else {
-			gotd_imsg_send_error(&ibuf, 0, PROC_SESSION_WRITE,
-			    err);
+			gotd_imsg_send_error(&ibuf, 0,
+			    GOTD_PROC_SESSION_WRITE, err);
 			imsgbuf_clear(&ibuf);
 		}
 	}
@@ -191,6 +198,140 @@ session_write_sighdlr(int sig, short event, void *arg)
 	}
 }
 
+static const struct got_error *
+recv_packfile_received(int *pack_empty, struct imsg *imsg)
+{
+	struct gotd_imsg_packfile_received recvd;
+	size_t datalen;
+
+	*pack_empty = 0;
+
+	log_debug("packfile-received received");
+
+	datalen = imsg->hdr.len - IMSG_HEADER_SIZE;
+	if (datalen != sizeof(recvd))
+		return got_error(GOT_ERR_PRIVSEP_LEN);
+	memcpy(&recvd, imsg->data, sizeof(recvd));
+
+	if (recvd.pack_empty)
+		*pack_empty = 1;
+
+	return NULL;
+}
+
+static const struct got_error *
+request_gotsys_conf(struct gotd_imsgev *iev)
+{
+	struct gotd_imsg_packfile_get_content content_req;
+	const char *refname = "refs/heads/main";
+	const char *path = "gotsys.conf";
+	struct ibuf *wbuf;
+	size_t len;
+	int fd = -1;
+
+	if (ftruncate(gotd_session.content_fd, 0L) == -1)
+		return got_error_from_errno("ftruncate");
+	
+	len = sizeof(content_req) + strlen(refname) + strlen(path);
+	wbuf = imsg_create(&iev->ibuf, GOTD_IMSG_PACKFILE_GET_CONTENT,
+	    GOTD_PROC_SESSION_WRITE, gotd_session.pid, len);
+	if (wbuf == NULL)
+		return got_error_from_errno("imsg_create PACKFILE_GET_CONTENT");
+
+	memset(&content_req, 0, sizeof(content_req));
+	content_req.refname_len = strlen(refname);
+	content_req.path_len = strlen(path);
+
+	if (imsg_add(wbuf, &content_req, sizeof(content_req)) == -1)
+		return got_error_from_errno("imsg_add PACKFILE_GET_CONTENT");
+	if (imsg_add(wbuf, refname, content_req.refname_len) == -1)
+		return got_error_from_errno("imsg_add PACKFILE_GET_CONTENT");
+	if (imsg_add(wbuf, path, content_req.path_len) == -1)
+		return got_error_from_errno("imsg_add PACKFILE_GET_CONTENT");
+
+	fd = dup(gotd_session.content_fd);
+	if (fd == -1) {
+		ibuf_free(wbuf);
+		return got_error_from_errno("dup");
+	}
+	ibuf_fd_set(wbuf, fd);
+
+	imsg_close(&iev->ibuf, wbuf);
+
+	return gotd_imsg_flush(&iev->ibuf);
+}
+
+static int
+need_packfile_verification(void)
+{
+	return (strcmp(gotd_session.repo_name, "gotsys") == 0 ||
+	    strcmp(gotd_session.repo_name, "gotsys.git") == 0);
+}
+
+static const struct got_error *
+verify_packfile(struct gotd_imsgev *iev)
+{
+	/* For now, verification is only implemented for gotsys.git. */
+	return request_gotsys_conf(iev);
+}
+
+static const struct got_error *
+recv_content_written(int *ref_found, struct imsg *imsg)
+{
+	struct gotd_imsg_packfile_content_written cw;
+	size_t datalen;
+
+	*ref_found = 0;
+
+	log_debug("content-written received");
+
+	datalen = imsg->hdr.len - IMSG_HEADER_SIZE;
+	if (datalen != sizeof(cw))
+		return got_error(GOT_ERR_PRIVSEP_LEN);
+	memcpy(&cw, imsg->data, sizeof(cw));
+	
+	if (cw.ref_found) {
+		*ref_found = 1;
+
+		/* Currently we only look for gotsys.conf content. */
+		if (!cw.wrote_content) {
+			return got_error_msg(GOT_ERR_BAD_OBJ_DATA,
+			    "gotsys.conf not found in pack file");
+		}
+	}
+
+	return NULL;
+}
+
+static const struct got_error *
+verify_gotsys_conf(void)
+{
+	struct gotd_imsgev *iev = &gotd_session.parent_iev;
+	int fd;
+
+	fd = dup(gotd_session.content_fd);
+	if (fd == -1)
+		return got_error_from_errno("dup");
+
+	if (gotd_imsg_compose_event(iev, GOTD_IMSG_RUN_GOTSYS_CHECK,
+	    GOTD_PROC_SESSION_WRITE, fd, NULL, 0) == -1) {
+		close(fd);
+		return got_error_from_errno("imsg compose RUN_GOTSYS_CHECK");
+	}
+
+	return NULL;
+}
+
+static const struct got_error *
+send_packfile_verified(struct gotd_imsgev *iev)
+{
+	if (gotd_imsg_compose_event(iev, GOTD_IMSG_PACKFILE_VERIFIED,
+	    GOTD_PROC_SESSION_WRITE, -1, NULL, 0) == -1)
+		return got_error_from_errno("imsg compose PACKFILE_VERIFIED");
+
+	return NULL;
+}
+
 static const struct got_error *
 recv_packfile_install(struct imsg *imsg)
 {
@@ -255,7 +396,7 @@ send_ref_update_ok(struct gotd_session_client *client,
 
 	len = sizeof(iok) + iok.name_len;
 	wbuf = imsg_create(&iev->ibuf, GOTD_IMSG_REF_UPDATE_OK,
-	    PROC_SESSION_WRITE, gotd_session.pid, len);
+	    GOTD_PROC_SESSION_WRITE, gotd_session.pid, len);
 	if (wbuf == NULL)
 		return got_error_from_errno("imsg_create REF_UPDATE_OK");
 
@@ -269,12 +410,14 @@ send_ref_update_ok(struct gotd_session_client *client,
 	return NULL;
 }
 
-static void
-send_refs_updated(struct gotd_session_client *client)
+static const struct got_error *
+send_refs_updated(struct gotd_imsgev *iev)
 {
-	if (gotd_imsg_compose_event(&client->iev, GOTD_IMSG_REFS_UPDATED,
-	    PROC_SESSION_WRITE, -1, NULL, 0) == -1)
-		log_warn("imsg compose REFS_UPDATED");
+	if (gotd_imsg_compose_event(iev, GOTD_IMSG_REFS_UPDATED,
+	    GOTD_PROC_SESSION_WRITE, -1, NULL, 0) == -1)
+		return got_error_from_errno("imsg compose REFS_UPDATED");
+
+	return NULL;
 }
 
 static const struct got_error *
@@ -298,7 +441,7 @@ send_ref_update_ng(struct gotd_session_client *client,
 
 	len = sizeof(ing) + ing.name_len + ing.reason_len;
 	wbuf = imsg_create(&iev->ibuf, GOTD_IMSG_REF_UPDATE_NG,
-	    PROC_SESSION_WRITE, gotd_session.pid, len);
+	    GOTD_PROC_SESSION_WRITE, gotd_session.pid, len);
 	if (wbuf == NULL)
 		return got_error_from_errno("imsg_create REF_UPDATE_NG");
 
@@ -419,23 +562,22 @@ queue_notification(struct got_object_id *old_id, struct got_object_id *new_id,
     struct got_repository *repo, struct got_reference *ref)
 {
 	const struct got_error *err = NULL;
-	struct gotd_repo *repo_cfg = gotd_session.repo_cfg;
 	struct gotd_imsgev *iev = &gotd_session.repo_child_iev;
 	struct got_pathlist_entry *pe;
 	struct gotd_session_notif *notif;
 
 	if (iev->ibuf.fd == -1 ||
-	    STAILQ_EMPTY(&repo_cfg->notification_targets))
+	    STAILQ_EMPTY(&gotd_session.notification_targets))
 		return NULL; /* notifications unused */
 
-	RB_FOREACH(pe, got_pathlist_head, &repo_cfg->notification_refs) {
+	RB_FOREACH(pe, got_pathlist_head, &gotd_session.notification_refs) {
 		const char *refname = pe->path;
 		if (strcmp(got_ref_get_name(ref), refname) == 0)
 			break;
 	}
 	if (pe == NULL) {
 		RB_FOREACH(pe, got_pathlist_head,
-		    &repo_cfg->notification_ref_namespaces) {
+		    &gotd_session.notification_ref_namespaces) {
 			const char *namespace = pe->path;
 
 			err = validate_namespace(namespace);
@@ -452,8 +594,8 @@ queue_notification(struct got_object_id *old_id, struct got_object_id *new_id,
 	 * configuration file then only send notifications if a match
 	 * was found.
 	 */
-	if (pe == NULL && (!RB_EMPTY(&repo_cfg->notification_refs) ||
-	    !RB_EMPTY(&repo_cfg->notification_ref_namespaces)))
+	if (pe == NULL && (!RB_EMPTY(&gotd_session.notification_refs) ||
+	    !RB_EMPTY(&gotd_session.notification_ref_namespaces)))
 		return NULL;
 
 	notif = calloc(1, sizeof(*notif));
@@ -573,16 +715,16 @@ forward_notification(struct gotd_session_client *client, struct imsg *imsg)
 		goto done;
 	}
 
-	strlcpy(inotify.repo_name, gotd_session.repo_cfg->name,
+	strlcpy(inotify.repo_name, gotd_session.repo_name,
 	    sizeof(inotify.repo_name));
 
 	snprintf(inotify.subject_line, sizeof(inotify.subject_line),
-	    "%s: %s %s %s: %.12s", gotd_session.repo_cfg->name,
+	    "%s: %s %s %s: %.12s", gotd_session.repo_name,
 	    client->username, action, notif->refname, id_str);
 
 	inotify.username_len = strlen(client->username);
 	wbuf = imsg_create(&iev->ibuf, GOTD_IMSG_NOTIFY,
-	    PROC_SESSION_WRITE, gotd_session.pid,
+	    GOTD_PROC_SESSION_WRITE, gotd_session.pid,
 	    sizeof(inotify) + inotify.username_len);
 	if (wbuf == NULL) {
 		err = got_error_from_errno("imsg_create NOTIFY");
@@ -635,7 +777,7 @@ request_notification(struct gotd_session_notif *notif)
 
 	len = sizeof(icontent) + icontent.refname_len;
 	wbuf = imsg_create(&iev->ibuf, GOTD_IMSG_NOTIFY,
-	    PROC_SESSION_WRITE, gotd_session.pid, len);
+	    GOTD_PROC_SESSION_WRITE, gotd_session.pid, len);
 	if (wbuf == NULL) {
 		err = got_error_from_errno("imsg_create NOTIFY");
 		goto done;
@@ -675,7 +817,6 @@ update_ref(int *shut, struct gotd_session_client *client,
 	struct got_reference *ref = NULL;
 	struct gotd_imsg_ref_update iref;
 	struct got_object_id old_id, new_id;
-	struct gotd_session_notif *notif;
 	struct got_object_id *id = NULL;
 	char *refname = NULL;
 	size_t datalen;
@@ -815,24 +956,6 @@ done:
 	} else
 		send_ref_update_ok(client, &iref, refname);
 
-	if (client->nref_updates > 0) {
-		client->nref_updates--;
-		if (client->nref_updates == 0) {
-			send_refs_updated(client);
-			notif = STAILQ_FIRST(&notifications);
-			if (notif) {
-				gotd_session.state = GOTD_STATE_NOTIFY;
-				err = request_notification(notif);
-				if (err) {
-					log_warn("could not send notification: "
-					    "%s", err->msg);
-					client->flush_disconnect = 1;
-				}
-			} else
-				client->flush_disconnect = 1;
-		}
-
-	}
 	if (locked) {
 		const struct got_error *unlock_err;
 		unlock_err = got_ref_unlock(ref);
@@ -892,6 +1015,9 @@ session_dispatch_repo_child(int fd, short event, void *arg)
 		uint32_t client_id = 0;
 		int do_disconnect = 0;
 		int do_ref_updates = 0, do_ref_update = 0;
+		int do_packfile_verification = 0;
+		int do_content_verification = 0;
+		int packfile_verified = 0;
 		int do_packfile_install = 0, do_notify = 0;
 
 		if ((n = imsg_get(ibuf, &imsg)) == -1)
@@ -904,6 +1030,30 @@ session_dispatch_repo_child(int fd, short event, void *arg)
 			do_disconnect = 1;
 			err = gotd_imsg_recv_error(&client_id, &imsg);
 			break;
+		case GOTD_IMSG_PACKFILE_RECEIVED: {
+			int pack_empty;
+
+			err = recv_packfile_received(&pack_empty, &imsg);
+			if (err)
+				break;
+			if (!pack_empty && need_packfile_verification())
+				do_packfile_verification = 1;
+			else
+				packfile_verified = 1;
+			break;
+		}
+		case GOTD_IMSG_PACKFILE_CONTENT_WRITTEN: {
+			int ref_found;
+
+			err = recv_content_written(&ref_found, &imsg);
+			if (err == NULL) {
+				if (ref_found)
+					do_content_verification = 1;
+				else
+					packfile_verified = 1;
+			}
+			break;
+		}
 		case GOTD_IMSG_PACKFILE_INSTALL:
 			err = recv_packfile_install(&imsg);
 			if (err == NULL)
@@ -937,7 +1087,13 @@ session_dispatch_repo_child(int fd, short event, void *arg)
 		} else {
 			struct gotd_session_notif *notif;
 
-			if (do_packfile_install)
+			if (do_packfile_verification) {
+				err = verify_packfile(iev);
+			} else if (do_content_verification) {
+				err = verify_gotsys_conf();
+			} else if (packfile_verified) {
+				err = send_packfile_verified(iev);
+			} else if (do_packfile_install)
 				err = install_pack(client,
 				    gotd_session.repo->path, &imsg);
 			else if (do_ref_updates)
@@ -950,6 +1106,18 @@ session_dispatch_repo_child(int fd, short event, void *arg)
 			if (err)
 				log_warnx("uid %d: %s", client->euid, err->msg);
 
+			if (do_ref_update && client->nref_updates > 0) {
+				client->nref_updates--;
+				if (client->nref_updates == 0) {
+					err = send_refs_updated(
+					    &gotd_session.parent_iev);
+					if (err) {
+						log_warn("%s", err->msg);
+						shut = 1;
+					}
+				}
+			}
+
 			notif = STAILQ_FIRST(&notifications);
 			if (notif && do_notify) {
 				/* Request content for next notification. */
@@ -1066,7 +1234,7 @@ forward_ref_update(struct gotd_session_client *client, struct imsg *imsg)
 	memcpy(iref, imsg->data, datalen);
 
 	if (gotd_imsg_compose_event(&gotd_session.repo_child_iev,
-	    GOTD_IMSG_REF_UPDATE, PROC_SESSION_WRITE, -1,
+	    GOTD_IMSG_REF_UPDATE, GOTD_PROC_SESSION_WRITE, -1,
 	    iref, datalen) == -1)
 		err = got_error_from_errno("imsg compose REF_UPDATE");
 	free(iref);
@@ -1110,19 +1278,20 @@ recv_packfile(struct gotd_session_client *client)
 
 	/* Send pack pipe end 0 to repo child process. */
 	if (gotd_imsg_compose_event(&gotd_session.repo_child_iev,
-	    GOTD_IMSG_PACKFILE_PIPE, PROC_SESSION_WRITE, pipe[0],
+	    GOTD_IMSG_PACKFILE_PIPE, GOTD_PROC_SESSION_WRITE, pipe[0],
 	        NULL, 0) == -1) {
 		err = got_error_from_errno("imsg compose PACKFILE_PIPE");
-		pipe[0] = -1;
 		goto done;
 	}
 	pipe[0] = -1;
 
 	/* Send pack pipe end 1 to gotsh(1) (expects just an fd, no data). */
 	if (gotd_imsg_compose_event(&client->iev,
-	    GOTD_IMSG_PACKFILE_PIPE, PROC_SESSION_WRITE, pipe[1],
-	    NULL, 0) == -1)
+	    GOTD_IMSG_PACKFILE_PIPE, GOTD_PROC_SESSION_WRITE, pipe[1],
+	    NULL, 0) == -1) {
 		err = got_error_from_errno("imsg compose PACKFILE_PIPE");
+		goto done;
+	}
 	pipe[1] = -1;
 
 	if (asprintf(&basepath, "%s/%s/receiving-from-uid-%d.pack",
@@ -1157,10 +1326,9 @@ recv_packfile(struct gotd_session_client *client)
 	}
 
 	if (gotd_imsg_compose_event(&gotd_session.repo_child_iev,
-	    GOTD_IMSG_PACKIDX_FILE, PROC_SESSION_WRITE,
+	    GOTD_IMSG_PACKIDX_FILE, GOTD_PROC_SESSION_WRITE,
 	    idxfd, NULL, 0) == -1) {
 		err = got_error_from_errno("imsg compose PACKIDX_FILE");
-		idxfd = -1;
 		goto done;
 	}
 	idxfd = -1;
@@ -1170,10 +1338,9 @@ recv_packfile(struct gotd_session_client *client)
 		ipack.report_status = 1;
 
 	if (gotd_imsg_compose_event(&gotd_session.repo_child_iev,
-	    GOTD_IMSG_RECV_PACKFILE, PROC_SESSION_WRITE, packfd,
+	    GOTD_IMSG_RECV_PACKFILE, GOTD_PROC_SESSION_WRITE, packfd,
 	    &ipack, sizeof(ipack)) == -1) {
 		err = got_error_from_errno("imsg compose RECV_PACKFILE");
-		packfd = -1;
 		goto done;
 	}
 	packfd = -1;
@@ -1379,7 +1546,7 @@ list_refs_request(void)
 		return got_error_from_errno("dup");
 
 	if (gotd_imsg_compose_event(iev, GOTD_IMSG_LIST_REFS_INTERNAL,
-	    PROC_SESSION_WRITE, fd, NULL, 0) == -1) {
+	    GOTD_PROC_SESSION_WRITE, fd, NULL, 0) == -1) {
 		err = got_error_from_errno("imsg compose LIST_REFS_INTERNAL");
 		close(fd);
 		return err;
@@ -1563,7 +1730,7 @@ recv_repo_child(struct imsg *imsg)
 
 	memcpy(&ichild, imsg->data, sizeof(ichild));
 
-	if (ichild.proc_id != PROC_REPO_WRITE)
+	if (ichild.proc_id != GOTD_PROC_REPO_WRITE)
 		return got_error_msg(GOT_ERR_PRIVSEP_MSG,
 		    "bad child process type");
 
@@ -1571,6 +1738,12 @@ recv_repo_child(struct imsg *imsg)
 	if (fd == -1)
 		return got_error(GOT_ERR_PRIVSEP_NO_FD);
 
+	if (strlcpy(gotd_session.repo_name, ichild.repo_name,
+	    sizeof(gotd_session.repo_name)) >= sizeof(gotd_session.repo_name)) {
+		return got_error_msg(GOT_ERR_NO_SPACE,
+		    "repository name too long");
+	}
+
 	if (imsgbuf_init(&gotd_session.repo_child_iev.ibuf, fd) == -1) {
 		close(fd);
 		return got_error_from_errno("imsgbuf_init");
@@ -1597,10 +1770,12 @@ session_dispatch(int fd, short event, void *arg)
 	const struct got_error *err = NULL;
 	struct gotd_imsgev *iev = arg;
 	struct imsgbuf *ibuf = &iev->ibuf;
+	struct gotd_imsgev *repo_child_iev = &gotd_session.repo_child_iev;
 	struct gotd_session_client *client = &gotd_session_client;
 	ssize_t n;
 	int shut = 0;
 	struct imsg imsg;
+	size_t npaths;
 
 	if (event & EV_READ) {
 		if ((n = imsgbuf_read(ibuf)) == -1)
@@ -1622,6 +1797,7 @@ session_dispatch(int fd, short event, void *arg)
 		const struct got_error *err = NULL;
 		uint32_t client_id = 0;
 		int do_disconnect = 0, do_list_refs = 0;
+		int send_notifications = 0;
 
 		if ((n = imsg_get(ibuf, &imsg)) == -1)
 			fatal("%s: imsg_get error", __func__);
@@ -1639,6 +1815,80 @@ session_dispatch(int fd, short event, void *arg)
 		case GOTD_IMSG_DISCONNECT:
 			do_disconnect = 1;
 			break;
+		case GOTD_IMSG_NOTIFICATION_REFS:
+			if (gotd_session.notification_refs_cur != NULL ||
+			    gotd_session.num_notification_refs_needed != 0) {
+				err = got_error(GOT_ERR_PRIVSEP_MSG);
+				break;
+			}
+			err = gotd_imsg_recv_pathlist(&npaths, &imsg);
+			if (err)
+				break;
+			gotd_session.notification_refs_cur =
+			    &gotd_session.notification_refs;
+			gotd_session.num_notification_refs_needed = npaths;
+			gotd_session.num_notification_refs_received = 0;
+			break;
+		case GOTD_IMSG_NOTIFICATION_REF_NAMESPACES:
+			if (gotd_session.notification_refs_cur != NULL ||
+			    gotd_session.num_notification_refs_needed != 0) {
+				err = got_error(GOT_ERR_PRIVSEP_MSG);
+				break;
+			}
+			err = gotd_imsg_recv_pathlist(&npaths, &imsg);
+			if (err)
+				break;
+			gotd_session.notification_refs_cur =
+			    &gotd_session.notification_ref_namespaces;
+			gotd_session.num_notification_refs_needed = npaths;
+			gotd_session.num_notification_refs_received = 0;
+			break;
+		case GOTD_IMSG_NOTIFICATION_REFS_ELEM:
+		case GOTD_IMSG_NOTIFICATION_REF_NAMESPACES_ELEM:
+			if (gotd_session.notification_refs_cur == NULL ||
+			    gotd_session.num_notification_refs_needed == 0 ||
+			    gotd_session.num_notification_refs_received >=
+			    gotd_session.num_notification_refs_needed) {
+				err = got_error(GOT_ERR_PRIVSEP_MSG);
+				break;
+			}
+			err = gotd_imsg_recv_pathlist_elem(&imsg,
+			    gotd_session.notification_refs_cur);
+			if (err)
+				break;
+			if (++gotd_session.num_notification_refs_received >=
+			    gotd_session.num_notification_refs_needed) {
+				gotd_session.notification_refs_cur = NULL;
+				gotd_session.num_notification_refs_needed = 0;
+			}
+			break;
+		case GOTD_IMSG_NOTIFICATION_TARGET_EMAIL: {
+			struct gotd_notification_target *target;
+
+			err = gotd_imsg_recv_notification_target_email(NULL,
+			    &target, &imsg);
+			if (err)
+				break;
+			STAILQ_INSERT_TAIL(&gotd_session.notification_targets,
+			    target, entry);
+			break;
+		}
+		case GOTD_IMSG_NOTIFICATION_TARGET_HTTP: {
+			struct gotd_notification_target *target;
+
+			err = gotd_imsg_recv_notification_target_http(NULL,
+			    &target, &imsg);
+			if (err)
+				break;
+			STAILQ_INSERT_TAIL(&gotd_session.notification_targets,
+			    target, entry);
+			break;
+		}
+		case GOTD_IMSG_REQUEST_TIMEOUT:
+			if (imsg_get_data(&imsg, &gotd_session.request_timeout,
+			    sizeof(gotd_session.request_timeout)) == -1)
+				err = got_error_from_errno("imsg_get_data");
+			break;
 		case GOTD_IMSG_CONNECT_NOTIFIER:
 			err = recv_notifier(&imsg);
 			break;
@@ -1648,6 +1898,27 @@ session_dispatch(int fd, short event, void *arg)
 				break;
 			do_list_refs = 1;
 			break;
+		case GOTD_IMSG_NOTIFY:
+			send_notifications = 1;
+			break;
+		case GOTD_IMSG_PACKFILE_VERIFIED:
+			if (gotd_session.state != GOTD_STATE_EXPECT_PACKFILE) {
+				err = got_error(GOT_ERR_PRIVSEP_MSG);
+				do_disconnect = 1;
+				break;
+			}
+			if (repo_child_iev->ibuf.fd == -1) {
+				err = got_error(GOT_ERR_PRIVSEP_MSG);
+				do_disconnect = 1;
+				break;
+			}
+			if (gotd_imsg_forward(repo_child_iev, &imsg,
+			    -1) == -1) {
+				err = got_error_from_errno("imsg compose "
+				    " PACKFILE_VERIFIED");
+				do_disconnect = 1;
+			}
+			break;
 		default:
 			log_debug("unexpected imsg %d", imsg.hdr.type);
 			break;
@@ -1661,6 +1932,27 @@ session_dispatch(int fd, short event, void *arg)
 				disconnect(client);
 		} else if (do_list_refs)
 			err = list_refs_request();
+		else if (send_notifications) {
+			struct gotd_session_notif *notif;
+
+			err = send_refs_updated(&client->iev);
+			if (err) {
+				log_warnx("uid %d: %s", client->euid, err->msg);
+				err = NULL;
+			}
+
+			notif = STAILQ_FIRST(&notifications);
+			if (notif) {
+				gotd_session.state = GOTD_STATE_NOTIFY;
+				err = request_notification(notif);
+				if (err) {
+					log_warn("could not send notification: "
+					    "%s", err->msg);
+					client->flush_disconnect = 1;
+				}
+			} else
+				client->flush_disconnect = 1;
+		}
 
 		if (err)
 			log_warnx("uid %d: %s", client->euid, err->msg);
@@ -1677,8 +1969,7 @@ done:
 
 void
 session_write_main(const char *title, const char *repo_path,
-    int *pack_fds, int *temp_fds, struct timeval *request_timeout,
-    struct gotd_repo *repo_cfg)
+    int *pack_fds, int *temp_fds, int content_fd)
 {
 	const struct got_error *err = NULL;
 	struct event evsigint, evsigterm, evsighup, evsigusr1;
@@ -1689,9 +1980,12 @@ session_write_main(const char *title, const char *repo_path,
 	gotd_session.pid = getpid();
 	gotd_session.pack_fds = pack_fds;
 	gotd_session.temp_fds = temp_fds;
-	memcpy(&gotd_session.request_timeout, request_timeout,
-	    sizeof(gotd_session.request_timeout));
-	gotd_session.repo_cfg = repo_cfg;
+	gotd_session.content_fd = content_fd;
+	gotd_session.request_timeout.tv_sec = GOTD_DEFAULT_REQUEST_TIMEOUT;
+	gotd_session.request_timeout.tv_usec = 0;
+	RB_INIT(&gotd_session.notification_refs);
+	RB_INIT(&gotd_session.notification_ref_namespaces);
+	STAILQ_INIT(&gotd_session.notification_targets);
 
 	if (imsgbuf_init(&gotd_session.notifier_iev.ibuf, -1) == -1) {
 		err = got_error_from_errno("imsgbuf_init");
@@ -1733,6 +2027,8 @@ session_write_main(const char *title, const char *repo_path,
 	gotd_session_client.delta_cache_fd = -1;
 	gotd_session_client.accept_flush_pkt = 1;
 
+	gotd_session.repo_child_iev.ibuf.fd = -1;
+
 	if (imsgbuf_init(&gotd_session.parent_iev.ibuf, GOTD_FILENO_MSG_PIPE)
 	    == -1) {
 		err = got_error_from_errno("imsgbuf_init");
@@ -1745,7 +2041,7 @@ session_write_main(const char *title, const char *repo_path,
 	event_set(&gotd_session.parent_iev.ev, gotd_session.parent_iev.ibuf.fd,
 	    EV_READ, session_dispatch, &gotd_session.parent_iev);
 	if (gotd_imsg_compose_event(&gotd_session.parent_iev,
-	    GOTD_IMSG_CLIENT_SESSION_READY, PROC_SESSION_WRITE,
+	    GOTD_IMSG_CLIENT_SESSION_READY, GOTD_PROC_SESSION_WRITE,
 	    -1, NULL, 0) == -1) {
 		err = got_error_from_errno("imsg compose CLIENT_SESSION_READY");
 		goto done;
