@@ -68,20 +68,11 @@ int
 config_getcfg(struct gotwebd *env, struct imsg *imsg)
 {
 	/* nothing to do but tell gotwebd configuration is done */
-	if (sockets_compose_main(env, IMSG_CFG_DONE, NULL, 0) == -1)
+	if (sockets_compose_main(env, GOTWEBD_IMSG_CFG_DONE, NULL, 0) == -1)
 		fatal("sockets_compose_main IMSG_CFG_DONE");
 	return 0;
 }
 
-int
-config_setserver(struct gotwebd *env, struct server *srv)
-{
-	if (main_compose_sockets(env, IMSG_CFG_SRV, -1, srv, sizeof(*srv))
-	    == -1)
-		fatal("main_compose_sockets IMSG_CFG_SRV");
-	return 0;
-}
-
 int
 config_getserver(struct gotwebd *env, struct imsg *imsg)
 {
@@ -106,17 +97,20 @@ config_getserver(struct gotwebd *env, struct imsg *imsg)
 }
 
 int
-config_setsock(struct gotwebd *env, struct socket *sock)
+config_setsock(struct gotwebd *env, struct socket *sock, uid_t uid, gid_t gid)
 {
 	/* open listening sockets */
-	if (sockets_privinit(env, sock) == -1)
+	if (sockets_privinit(env, sock, uid, gid) == -1)
 		return -1;
 
-	if (main_compose_sockets(env, IMSG_CFG_SOCK, sock->fd,
+	if (main_compose_sockets(env, GOTWEBD_IMSG_CFG_SOCK, sock->fd,
 	    &sock->conf, sizeof(sock->conf)) == -1)
 		fatal("main_compose_sockets IMSG_CFG_SOCK");
 
-	sock->fd = -1;
+	if (main_compose_gotweb(env, GOTWEBD_IMSG_CFG_SOCK, sock->fd,
+	    &sock->conf, sizeof(sock->conf)) == -1)
+		fatal("main_compose_gotweb GOTWEBD_IMSG_CFG_SOCK");
+
 	return 0;
 }
 
@@ -172,13 +166,13 @@ config_setfd(struct gotwebd *env)
 			fd = got_opentempfd();
 			if (fd == -1)
 				fatal("got_opentemp");
-			if (imsg_compose_event(&env->iev_server[j],
-			    IMSG_CFG_FD, 0, -1, fd, NULL, 0) == -1)
-				fatal("imsg_compose_event IMSG_CFG_FD");
+			if (imsg_compose_event(&env->iev_gotweb[j],
+			    GOTWEBD_IMSG_CFG_FD, 0, -1, fd, NULL, 0) == -1)
+				fatal("imsg_compose_event GOTWEBD_IMSG_CFG_FD");
 
-			if (imsgbuf_flush(&env->iev_server[j].ibuf) == -1)
+			if (imsgbuf_flush(&env->iev_gotweb[j].ibuf) == -1)
 				fatal("imsgbuf_flush");
-			imsg_event_add(&env->iev_server[j]);
+			imsg_event_add(&env->iev_gotweb[j]);
 		}
 	}
 
