@@ -1711,72 +1711,75 @@ struct repo_commit *rc = TAILQ_FIRST(&t->repo_commits);
 struct tm tm;
 char buf[BUFSIZ], datebuf[64];
 size_t r;
-#line 1242 "../gotwebd/pages.tmpl"
+int w;
+#line 1243 "../gotwebd/pages.tmpl"
 if (gmtime_r(&rc->committer_time, &tm) == NULL ||
 asctime_r(&tm, datebuf) == NULL)
 return (-1);
-#line 1246 "../gotwebd/pages.tmpl"
+#line 1247 "../gotwebd/pages.tmpl"
 datebuf[strcspn(datebuf, "\n")] = '\0';
-#line 1248 "../gotwebd/pages.tmpl"
+#line 1249 "../gotwebd/pages.tmpl"
 if ((tp_ret = tp_write(tp, "commit ", 7)) == -1) goto err;
-#line 1248 "../gotwebd/pages.tmpl"
+#line 1249 "../gotwebd/pages.tmpl"
 if ((tp_ret = tp_htmlescape(tp, rc->commit_id)) == -1)
 goto err;
-#line 1248 "../gotwebd/pages.tmpl"
+#line 1249 "../gotwebd/pages.tmpl"
 if ((tp_ret = tp_htmlescape(tp, "\n")) == -1)
 goto err;
 if ((tp_ret = tp_write(tp, "from: ", 6)) == -1) goto err;
-#line 1249 "../gotwebd/pages.tmpl"
+#line 1250 "../gotwebd/pages.tmpl"
 if ((tp_ret = tp_writes(tp, rc->author)) == -1)
 goto err;
-#line 1249 "../gotwebd/pages.tmpl"
+#line 1250 "../gotwebd/pages.tmpl"
 if ((tp_ret = tp_htmlescape(tp, "\n")) == -1)
 goto err;
 if (strcmp(rc->committer, rc->author) != 0) {
 if ((tp_ret = tp_write(tp, "via: ", 5)) == -1) goto err;
-#line 1251 "../gotwebd/pages.tmpl"
+#line 1252 "../gotwebd/pages.tmpl"
 if ((tp_ret = tp_writes(tp, rc->committer)) == -1)
 goto err;
-#line 1251 "../gotwebd/pages.tmpl"
+#line 1252 "../gotwebd/pages.tmpl"
 if ((tp_ret = tp_htmlescape(tp, "\n")) == -1)
 goto err;
 }
-#line 1253 "../gotwebd/pages.tmpl"
+#line 1254 "../gotwebd/pages.tmpl"
 if ((tp_ret = tp_write(tp, "date: ", 6)) == -1) goto err;
-#line 1253 "../gotwebd/pages.tmpl"
+#line 1254 "../gotwebd/pages.tmpl"
 if ((tp_ret = tp_htmlescape(tp, datebuf)) == -1)
 goto err;
-#line 1253 "../gotwebd/pages.tmpl"
+#line 1254 "../gotwebd/pages.tmpl"
 if ((tp_ret = tp_htmlescape(tp, " UTC")) == -1)
 goto err;
-#line 1253 "../gotwebd/pages.tmpl"
+#line 1254 "../gotwebd/pages.tmpl"
 if ((tp_ret = tp_htmlescape(tp, "\n")) == -1)
 goto err;
 if ((tp_ret = tp_htmlescape(tp, "\n")) == -1)
 goto err;
 if ((tp_ret = tp_writes(tp, rc->commit_msg)) == -1)
 goto err;
-#line 1255 "../gotwebd/pages.tmpl"
+#line 1256 "../gotwebd/pages.tmpl"
 if ((tp_ret = tp_htmlescape(tp, "\n")) == -1)
 goto err;
-#line 1257 "../gotwebd/pages.tmpl"
+#line 1258 "../gotwebd/pages.tmpl"
 if (template_flush(tp) == -1)
 return (-1);
 for (;;) {
 r = fread(buf, 1, sizeof(buf), t->fp);
-if (fcgi_write(c, buf, r) == -1 ||
-r != sizeof(buf))
+if (r == 0)
+break;
+w = fcgi_write(c, buf, r);
+if (w == -1)
 break;
 }
 err:
 return tp_ret;
 }
-#line 1268 "../gotwebd/pages.tmpl"
+#line 1271 "../gotwebd/pages.tmpl"
 int
 gotweb_render_rss(struct template *tp)
 {
 int tp_ret = 0;
-#line 1270 "../gotwebd/pages.tmpl"
+#line 1273 "../gotwebd/pages.tmpl"
 struct request *c = tp->tp_arg;
 struct server *srv = c->srv;
 struct transport *t = c->t;
@@ -1787,40 +1790,40 @@ struct gotweb_url summary = {
 .index_page = -1,
 .path = repo_dir->name,
 };
-#line 1284 "../gotwebd/pages.tmpl"
+#line 1287 "../gotwebd/pages.tmpl"
 if ((tp_ret = tp_write(tp, "<?xml version=\"1.0\" encoding=\"UTF-8\"?><rss version=\"2.0\" xmlns:content=\"http://purl.org/rss/1.0/modules/content/\"><channel><title>Tags of ", 138)) == -1) goto err;
-#line 1284 "../gotwebd/pages.tmpl"
+#line 1287 "../gotwebd/pages.tmpl"
 if ((tp_ret = tp_htmlescape(tp, repo_dir->name)) == -1)
 goto err;
-#line 1287 "../gotwebd/pages.tmpl"
+#line 1290 "../gotwebd/pages.tmpl"
 if ((tp_ret = tp_write(tp, "</title><link><![CDATA[", 23)) == -1) goto err;
-#line 1287 "../gotwebd/pages.tmpl"
+#line 1290 "../gotwebd/pages.tmpl"
 if ((tp_ret = gotweb_render_absolute_url(c, &summary)) == -1) goto err;
-#line 1290 "../gotwebd/pages.tmpl"
+#line 1293 "../gotwebd/pages.tmpl"
 if ((tp_ret = tp_write(tp, "]]></link>", 10)) == -1) goto err;
-#line 1290 "../gotwebd/pages.tmpl"
+#line 1293 "../gotwebd/pages.tmpl"
 if (srv->show_repo_description) {
 if ((tp_ret = tp_write(tp, "<description>", 13)) == -1) goto err;
-#line 1291 "../gotwebd/pages.tmpl"
+#line 1294 "../gotwebd/pages.tmpl"
 if ((tp_ret = tp_htmlescape(tp, repo_dir->description)) == -1)
 goto err;
 if ((tp_ret = tp_write(tp, "</description>", 14)) == -1) goto err;
 }
 TAILQ_FOREACH(rt, &t->repo_tags, entry) {
-#line 1294 "../gotwebd/pages.tmpl"
+#line 1297 "../gotwebd/pages.tmpl"
 if ((tp_ret = rss_tag_item(tp, rt)) == -1) goto err;
 }
-#line 1298 "../gotwebd/pages.tmpl"
+#line 1301 "../gotwebd/pages.tmpl"
 if ((tp_ret = tp_write(tp, "</channel></rss>", 16)) == -1) goto err;
 err:
 return tp_ret;
 }
-#line 1300 "../gotwebd/pages.tmpl"
+#line 1303 "../gotwebd/pages.tmpl"
 int
 rss_tag_item(struct template *tp, struct repo_tag *rt)
 {
 int tp_ret = 0;
-#line 1302 "../gotwebd/pages.tmpl"
+#line 1305 "../gotwebd/pages.tmpl"
 struct request *c = tp->tp_arg;
 struct transport *t = c->t;
 struct repo_dir *repo_dir = t->repo_dir;
@@ -1834,92 +1837,92 @@ struct gotweb_url tag = {
 .path = repo_dir->name,
 .commit = rt->commit_id,
 };
-#line 1316 "../gotwebd/pages.tmpl"
+#line 1319 "../gotwebd/pages.tmpl"
 if (strncmp(tag_name, "refs/tags/", 10) == 0)
 tag_name += 10;
-#line 1319 "../gotwebd/pages.tmpl"
+#line 1322 "../gotwebd/pages.tmpl"
 if (gmtime_r(&rt->tagger_time, &tm) == NULL)
 return -1;
 r = strftime(rfc822, sizeof(rfc822), "%a, %d %b %Y %H:%M:%S GMT", &tm);
 if (r == 0)
 return 0;
-#line 1326 "../gotwebd/pages.tmpl"
+#line 1329 "../gotwebd/pages.tmpl"
 if ((tp_ret = tp_write(tp, "<item><title>", 13)) == -1) goto err;
-#line 1326 "../gotwebd/pages.tmpl"
+#line 1329 "../gotwebd/pages.tmpl"
 if ((tp_ret = tp_htmlescape(tp, repo_dir->name)) == -1)
 goto err;
-#line 1326 "../gotwebd/pages.tmpl"
+#line 1329 "../gotwebd/pages.tmpl"
 if ((tp_ret = tp_htmlescape(tp, " ")) == -1)
 goto err;
-#line 1326 "../gotwebd/pages.tmpl"
+#line 1329 "../gotwebd/pages.tmpl"
 if ((tp_ret = tp_htmlescape(tp, tag_name)) == -1)
 goto err;
-#line 1329 "../gotwebd/pages.tmpl"
+#line 1332 "../gotwebd/pages.tmpl"
 if ((tp_ret = tp_write(tp, "</title><link><![CDATA[", 23)) == -1) goto err;
-#line 1329 "../gotwebd/pages.tmpl"
+#line 1332 "../gotwebd/pages.tmpl"
 if ((tp_ret = gotweb_render_absolute_url(c, &tag)) == -1) goto err;
-#line 1333 "../gotwebd/pages.tmpl"
+#line 1336 "../gotwebd/pages.tmpl"
 if ((tp_ret = tp_write(tp, "]]></link><description><![CDATA[<pre>", 37)) == -1) goto err;
-#line 1333 "../gotwebd/pages.tmpl"
+#line 1336 "../gotwebd/pages.tmpl"
 if ((tp_ret = tp_htmlescape(tp, rt->tag_commit)) == -1)
 goto err;
-#line 1335 "../gotwebd/pages.tmpl"
+#line 1338 "../gotwebd/pages.tmpl"
 if ((tp_ret = tp_write(tp, "</pre>]]></description>", 23)) == -1) goto err;
-#line 1335 "../gotwebd/pages.tmpl"
+#line 1338 "../gotwebd/pages.tmpl"
 if ((tp_ret = rss_author(tp, rt->tagger)) == -1) goto err;
 if ((tp_ret = tp_write(tp, "<guid isPermaLink=\"false\">", 26)) == -1) goto err;
-#line 1336 "../gotwebd/pages.tmpl"
+#line 1339 "../gotwebd/pages.tmpl"
 if ((tp_ret = tp_htmlescape(tp, rt->commit_id)) == -1)
 goto err;
-#line 1338 "../gotwebd/pages.tmpl"
+#line 1341 "../gotwebd/pages.tmpl"
 if ((tp_ret = tp_write(tp, "</guid><pubDate>", 16)) == -1) goto err;
-#line 1338 "../gotwebd/pages.tmpl"
+#line 1341 "../gotwebd/pages.tmpl"
 if ((tp_ret = tp_htmlescape(tp, rfc822)) == -1)
 goto err;
-#line 1341 "../gotwebd/pages.tmpl"
+#line 1344 "../gotwebd/pages.tmpl"
 if ((tp_ret = tp_write(tp, "</pubDate></item>", 17)) == -1) goto err;
 err:
 return tp_ret;
 }
-#line 1343 "../gotwebd/pages.tmpl"
+#line 1346 "../gotwebd/pages.tmpl"
 int
 rss_author(struct template *tp, char *author)
 {
 int tp_ret = 0;
-#line 1345 "../gotwebd/pages.tmpl"
+#line 1348 "../gotwebd/pages.tmpl"
 char *t, *mail;
-#line 1347 "../gotwebd/pages.tmpl"
+#line 1350 "../gotwebd/pages.tmpl"
 /* what to do if the author name contains a paren? */
 if (strchr(author, '(') != NULL || strchr(author, ')') != NULL)
 return 0;
-#line 1351 "../gotwebd/pages.tmpl"
+#line 1354 "../gotwebd/pages.tmpl"
 t = strchr(author, '<');
 if (t == NULL)
 return 0;
 *t = '\0';
 mail = t+1;
-#line 1357 "../gotwebd/pages.tmpl"
+#line 1360 "../gotwebd/pages.tmpl"
 while (isspace((unsigned char)*--t))
 *t = '\0';
-#line 1360 "../gotwebd/pages.tmpl"
+#line 1363 "../gotwebd/pages.tmpl"
 t = strchr(mail, '>');
 if (t == NULL)
 return 0;
 *t = '\0';
-#line 1366 "../gotwebd/pages.tmpl"
+#line 1369 "../gotwebd/pages.tmpl"
 if ((tp_ret = tp_write(tp, "<author>", 8)) == -1) goto err;
-#line 1366 "../gotwebd/pages.tmpl"
+#line 1369 "../gotwebd/pages.tmpl"
 if ((tp_ret = tp_htmlescape(tp, mail)) == -1)
 goto err;
-#line 1366 "../gotwebd/pages.tmpl"
+#line 1369 "../gotwebd/pages.tmpl"
 if ((tp_ret = tp_htmlescape(tp, " ")) == -1)
 goto err;
-#line 1366 "../gotwebd/pages.tmpl"
+#line 1369 "../gotwebd/pages.tmpl"
 if ((tp_ret = tp_write(tp, "(", 1)) == -1) goto err;
-#line 1366 "../gotwebd/pages.tmpl"
+#line 1369 "../gotwebd/pages.tmpl"
 if ((tp_ret = tp_htmlescape(tp, author)) == -1)
 goto err;
-#line 1368 "../gotwebd/pages.tmpl"
+#line 1371 "../gotwebd/pages.tmpl"
 if ((tp_ret = tp_write(tp, ")</author>", 10)) == -1) goto err;
 err:
 return tp_ret;
