@@ -49,7 +49,7 @@ struct got_fileindex;
  * will be created. The path itself must not yet exist, but the dirname(3)
  * of the path must already exist.
  * The reference provided will be used to determine the new worktree's
- * base commit. The third argument speficies the work tree's path prefix.
+ * base commit. The third argument specifies the work tree's path prefix.
  * The fourth argument specifies the meta data directory to use, which
  * should be either GOT_WORKTREE_GOT_DIR or GOT_WORKTREE_CVG_DIR.
  */
@@ -90,6 +90,11 @@ const char *got_worktree_get_path_prefix(struct got_worktree *);
  */
 const struct got_error *got_worktree_get_uuid(char **, struct got_worktree *);
 
+/*
+ * Get the work tree's format version number.
+ */
+int got_worktree_get_format_version(struct got_worktree *);
+
 /*
  * Check if a user-provided path prefix matches that of the worktree.
  */
@@ -284,10 +289,10 @@ const struct got_error *got_worktree_commit(struct got_object_id **,
     const char *, int, int, int, got_worktree_commit_msg_cb, void *,
     got_worktree_status_cb, void *, struct got_repository *);
 
-/* Get the path of a commitable worktree item. */
+/* Get the path of a committable worktree item. */
 const char *got_commitable_get_path(struct got_commitable *);
 
-/* Get the status of a commitable worktree item. */
+/* Get the status of a committable worktree item. */
 unsigned int got_commitable_get_status(struct got_commitable *);
 
 /*
@@ -578,6 +583,21 @@ const struct got_error *got_worktree_unstage(struct got_worktree *,
     struct got_pathlist_head *, got_worktree_checkout_cb, void *,
     got_worktree_patch_cb, void *, struct got_repository *);
 
+/*
+ * Prepare for getting meta data for paths in the work tree.  This
+ * function also returns a pointer to a fileindex which must be passed
+ * back to other path_info-related functions and *_version() functions.
+ */
+const struct got_error *
+got_worktree_path_info_prepare(struct got_fileindex **,
+    struct got_worktree *, struct got_repository *);
+
+/*
+ * Get the file-index version.
+ */
+uint32_t
+got_worktree_get_fileindex_version(struct got_fileindex *);
+
 /* A callback function which is invoked with per-path info. */
 typedef const struct got_error *(*got_worktree_path_info_cb)(void *,
     const char *path, mode_t mode, time_t mtime,
@@ -590,8 +610,15 @@ typedef const struct got_error *(*got_worktree_path_info_cb)(void *,
  * a path, and meta-data arguments (see got_worktree_path_info_cb).
  */
 const struct got_error *
-got_worktree_path_info(struct got_worktree *, struct got_pathlist_head *,
-    got_worktree_path_info_cb, void *, got_cancel_cb , void *);
+got_worktree_path_info(struct got_worktree *, struct got_fileindex *,
+    struct got_pathlist_head *, got_worktree_path_info_cb, void *,
+    got_cancel_cb , void *);
+
+/*
+ * Complete the current path_info operation.
+ */
+const struct got_error *
+got_worktree_path_info_complete(struct got_fileindex *, struct got_worktree *);
 
 /* References pointing at pre-rebase commit backups. */
 #define GOT_WORKTREE_REBASE_BACKUP_REF_PREFIX "refs/got/backup/rebase"
@@ -604,7 +631,7 @@ got_worktree_path_info(struct got_worktree *, struct got_pathlist_head *,
  */
 const struct got_error *
 got_worktree_patch_prepare(struct got_fileindex **, char **,
-    struct got_worktree *);
+    struct got_worktree *, struct got_repository *);
 
 /*
  * Lookup paths for the "old" and "new" file before patching and check their
@@ -626,4 +653,5 @@ got_worktree_patch_schedule_rm(const char *, struct got_repository *,
 
 /* Complete the patch operation. */
 const struct got_error *
-got_worktree_patch_complete(struct got_fileindex *, const char *);
+got_worktree_patch_complete(struct got_worktree *, struct got_fileindex *,
+    const char *);
