@@ -115,7 +115,7 @@ got_object_get_path(char **path, struct got_object_id *id,
 		goto done;
 
 	if (asprintf(path, "%s/%.2x/%s", path_objects,
-	    id->sha1[0], hex + 2) == -1)
+	    id->hash[0], hex + 2) == -1)
 		err = got_error_from_errno("asprintf");
 
 done:
@@ -136,7 +136,10 @@ got_object_open_loose_fd(int *fd, struct got_object_id *id,
 		return err;
 	*fd = open(path, O_RDONLY | O_NOFOLLOW | O_CLOEXEC);
 	if (*fd == -1) {
-		err = got_error_from_errno2("open", path);
+		if (errno == ENOENT)
+			err = got_error_no_obj(id);
+		else
+			err = got_error_from_errno2("open", path);
 		goto done;
 	}
 done:
@@ -150,7 +153,7 @@ got_object_open_by_id_str(struct got_object **obj, struct got_repository *repo,
 {
 	struct got_object_id id;
 
-	if (!got_parse_object_id(&id, id_str, GOT_HASH_SHA1))
+	if (!got_parse_object_id(&id, id_str, repo->algo))
 		return got_error_path(id_str, GOT_ERR_BAD_OBJ_ID_STR);
 
 	return got_object_open(obj, repo, &id);
