@@ -253,7 +253,7 @@ got_object_read_raw(uint8_t **outbuf, off_t *size, size_t *hdrlen,
 	*size = 0;
 	*hdrlen = 0;
 
-	got_hash_init(&ctx, GOT_HASH_SHA1);
+	got_hash_init(&ctx, expected_id->algo);
 	memset(&csum, 0, sizeof(csum));
 	csum.output_ctx = &ctx;
 
@@ -282,7 +282,6 @@ got_object_read_raw(uint8_t **outbuf, off_t *size, size_t *hdrlen,
 		f = fdopen(fd, "r");
 		if (f == NULL) {
 			err = got_error_from_errno("fdopen");
-			abort();
 			close(fd);
 			goto done;
 		}
@@ -332,7 +331,7 @@ got_object_commit_alloc_partial(void)
 
 const struct got_error *
 got_object_commit_add_parent(struct got_commit_object *commit,
-    const char *id_str)
+    const char *id_str, enum got_hash_algorithm algo)
 {
 	const struct got_error *err = NULL;
 	struct got_object_qid *qid;
@@ -341,7 +340,7 @@ got_object_commit_add_parent(struct got_commit_object *commit,
 	if (err)
 		return err;
 
-	if (!got_parse_object_id(&qid->id, id_str, GOT_HASH_SHA1)) {
+	if (!got_parse_object_id(&qid->id, id_str, algo)) {
 		err = got_error(GOT_ERR_BAD_OBJ_DATA);
 		got_object_qid_free(qid);
 		return err;
@@ -571,14 +570,15 @@ got_object_commit_get_logmsg_raw(struct got_commit_object *commit)
 
 const struct got_error *
 got_object_parse_commit(struct got_commit_object **commit, char *buf,
-    size_t len)
+    size_t len, enum got_hash_algorithm algo)
 {
 	const struct got_error *err = NULL;
-	enum got_hash_algorithm algo = GOT_HASH_SHA1;
 	char *s = buf;
-	size_t label_len;
+	size_t label_len, digest_string_len;
 	ssize_t remain = (ssize_t)len;
 
+	digest_string_len = got_hash_digest_string_length(algo);
+
 	if (remain == 0)
 		return got_error(GOT_ERR_BAD_OBJ_DATA);
 
@@ -589,7 +589,7 @@ got_object_parse_commit(struct got_commit_object **commit, char *buf,
 	label_len = strlen(GOT_COMMIT_LABEL_TREE);
 	if (strncmp(s, GOT_COMMIT_LABEL_TREE, label_len) == 0) {
 		remain -= label_len;
-		if (remain < SHA1_DIGEST_STRING_LENGTH) {
+		if (remain < digest_string_len) {
 			err = got_error(GOT_ERR_BAD_OBJ_DATA);
 			goto done;
 		}
@@ -598,8 +598,8 @@ got_object_parse_commit(struct got_commit_object **commit, char *buf,
 			err = got_error(GOT_ERR_BAD_OBJ_DATA);
 			goto done;
 		}
-		remain -= SHA1_DIGEST_STRING_LENGTH;
-		s += SHA1_DIGEST_STRING_LENGTH;
+		remain -= digest_string_len;
+		s += digest_string_len;
 	} else {
 		err = got_error(GOT_ERR_BAD_OBJ_DATA);
 		goto done;
@@ -608,17 +608,17 @@ got_object_parse_commit(struct got_commit_object **commit, char *buf,
 	label_len = strlen(GOT_COMMIT_LABEL_PARENT);
 	while (strncmp(s, GOT_COMMIT_LABEL_PARENT, label_len) == 0) {
 		remain -= label_len;
-		if (remain < SHA1_DIGEST_STRING_LENGTH) {
+		if (remain < digest_string_len) {
 			err = got_error(GOT_ERR_BAD_OBJ_DATA);
 			goto done;
 		}
 		s += label_len;
-		err = got_object_commit_add_parent(*commit, s);
+		err = got_object_commit_add_parent(*commit, s, algo);
 		if (err)
 			goto done;
 
-		remain -= SHA1_DIGEST_STRING_LENGTH;
-		s += SHA1_DIGEST_STRING_LENGTH;
+		remain -= digest_string_len;
+		s += digest_string_len;
 	}
 
 	label_len = strlen(GOT_COMMIT_LABEL_AUTHOR);
@@ -708,7 +708,7 @@ got_object_read_commit(struct got_commit_object **commit, int fd,
 	struct got_hash ctx;
 	struct got_object_id id;
 
-	got_hash_init(&ctx, GOT_HASH_SHA1);
+	got_hash_init(&ctx, expected_id->algo);
 	memset(&csum, 0, sizeof(csum));
 	csum.output_ctx = &ctx;
 
@@ -738,7 +738,8 @@ got_object_read_commit(struct got_commit_object **commit, int fd,
 
 	/* Skip object header. */
 	len -= obj->hdrlen;
-	err = got_object_parse_commit(commit, p + obj->hdrlen, len);
+	err = got_object_parse_commit(commit, p + obj->hdrlen, len,
+	    expected_id->algo);
 done:
 	free(p);
 	if (obj)
@@ -761,7 +762,7 @@ got_object_tree_close(struct got_tree_object *tree)
 
 const struct got_error *
 got_object_parse_tree_entry(struct got_parsed_tree_entry *pte, size_t *elen,
-    char *buf, size_t maxlen)
+    char *buf, size_t maxlen, size_t digest_len, enum got_hash_algorithm algo)
 {
 	char *p, *space;
 
@@ -785,14 +786,16 @@ got_object_parse_tree_entry(struct got_parsed_tree_entry *pte, size_t *elen,
 		p++;
 	}
 
-	if (*elen > maxlen || maxlen - *elen < SHA1_DIGEST_LENGTH)
+	if (*elen > maxlen || maxlen - *elen < digest_len)
 		return got_error(GOT_ERR_BAD_OBJ_DATA);
 
 	pte->name = space + 1;
 	pte->namelen = strlen(pte->name);
 	buf += *elen;
 	pte->id = buf;
-	*elen += SHA1_DIGEST_LENGTH;
+	pte->digest_len = digest_len;
+	pte->algo = algo;
+	*elen += digest_len;
 	return NULL;
 }
 
@@ -806,14 +809,17 @@ pte_cmp(const void *pa, const void *pb)
 
 const struct got_error *
 got_object_parse_tree(struct got_parsed_tree_entry **entries, size_t *nentries,
-    size_t *nentries_alloc, uint8_t *buf, size_t len)
+    size_t *nentries_alloc, uint8_t *buf, size_t len,
+    enum got_hash_algorithm algo)
 {
 	const struct got_error *err = NULL;
-	size_t remain = len;
+	size_t digest_len, remain = len;
 	const size_t nalloc = 16;
 	struct got_parsed_tree_entry *pte;
 	int i;
 
+	digest_len = got_hash_digest_length(algo);
+
 	*nentries = 0;
 	if (remain == 0)
 		return NULL; /* tree is empty */
@@ -833,7 +839,8 @@ got_object_parse_tree(struct got_parsed_tree_entry **entries, size_t *nentries,
 		}
 
 		pte = &(*entries)[*nentries];
-		err = got_object_parse_tree_entry(pte, &elen, buf, remain);
+		err = got_object_parse_tree_entry(pte, &elen, buf, remain,
+		    digest_len, algo);
 		if (err)
 			goto done;
 		buf += elen;
@@ -877,7 +884,7 @@ got_object_read_tree(struct got_parsed_tree_entry **entries, size_t *nentries,
 	struct got_hash ctx;
 	struct got_object_id id;
 
-	got_hash_init(&ctx, GOT_HASH_SHA1);
+	got_hash_init(&ctx, expected_id->algo);
 	memset(&csum, 0, sizeof(csum));
 	csum.output_ctx = &ctx;
 
@@ -903,7 +910,7 @@ got_object_read_tree(struct got_parsed_tree_entry **entries, size_t *nentries,
 	/* Skip object header. */
 	len -= obj->hdrlen;
 	err = got_object_parse_tree(entries, nentries, nentries_alloc,
-	    *p + obj->hdrlen, len);
+	    *p + obj->hdrlen, len, expected_id->algo);
 done:
 	if (obj)
 		got_object_close(obj);
@@ -926,13 +933,15 @@ got_object_tag_close(struct got_tag_object *tag)
 }
 
 const struct got_error *
-got_object_parse_tag(struct got_tag_object **tag, uint8_t *buf, size_t len)
+got_object_parse_tag(struct got_tag_object **tag, uint8_t *buf, size_t len,
+    enum got_hash_algorithm algo)
 {
 	const struct got_error *err = NULL;
-	enum got_hash_algorithm algo = GOT_HASH_SHA1;
 	size_t remain = len;
 	char *s = buf;
-	size_t label_len;
+	size_t label_len, digest_string_len;
+
+	digest_string_len = got_hash_digest_string_length(algo);
 
 	if (remain == 0)
 		return got_error(GOT_ERR_BAD_OBJ_DATA);
@@ -944,7 +953,7 @@ got_object_parse_tag(struct got_tag_object **tag, uint8_t *buf, size_t len)
 	label_len = strlen(GOT_TAG_LABEL_OBJECT);
 	if (strncmp(s, GOT_TAG_LABEL_OBJECT, label_len) == 0) {
 		remain -= label_len;
-		if (remain < SHA1_DIGEST_STRING_LENGTH) {
+		if (remain < digest_string_len) {
 			err = got_error(GOT_ERR_BAD_OBJ_DATA);
 			goto done;
 		}
@@ -953,8 +962,8 @@ got_object_parse_tag(struct got_tag_object **tag, uint8_t *buf, size_t len)
 			err = got_error(GOT_ERR_BAD_OBJ_DATA);
 			goto done;
 		}
-		remain -= SHA1_DIGEST_STRING_LENGTH;
-		s += SHA1_DIGEST_STRING_LENGTH;
+		remain -= digest_string_len;
+		s += digest_string_len;
 	} else {
 		err = got_error(GOT_ERR_BAD_OBJ_DATA);
 		goto done;
@@ -1117,7 +1126,7 @@ got_object_read_tag(struct got_tag_object **tag, int fd,
 	struct got_hash ctx;
 	struct got_object_id id;
 
-	got_hash_init(&ctx, GOT_HASH_SHA1);
+	got_hash_init(&ctx, expected_id->algo);
 	memset(&csum, 0, sizeof(csum));
 	csum.output_ctx = &ctx;
 
@@ -1143,7 +1152,8 @@ got_object_read_tag(struct got_tag_object **tag, int fd,
 
 	/* Skip object header. */
 	len -= obj->hdrlen;
-	err = got_object_parse_tag(tag, p + obj->hdrlen, len);
+	err = got_object_parse_tag(tag, p + obj->hdrlen, len,
+	    expected_id->algo);
 done:
 	free(p);
 	if (obj)
