@@ -1727,79 +1727,6 @@ got_repo_unpin_pack(struct got_repository *repo)
 	repo->pinned_pid = 0;
 }
 
-const struct got_error *
-got_repo_init(const char *repo_path, const char *head_name,
-    enum got_hash_algorithm algo)
-{
-	const struct got_error *err = NULL;
-	const char *dirnames[] = {
-		GOT_OBJECTS_DIR,
-		GOT_OBJECTS_PACK_DIR,
-		GOT_REFS_DIR,
-	};
-	const char *description_str = "Unnamed repository; "
-	    "edit this file 'description' to name the repository.";
-	const char *headref = "ref: refs/heads/";
-	const char *gitconfig_sha1 = "[core]\n"
-	    "\trepositoryformatversion = 0\n"
-	    "\tfilemode = true\n"
-	    "\tbare = true\n";
-	const char *gitconfig_sha256 = "[core]\n"
-	    "\trepositoryformatversion = 1\n"
-	    "\tfilemode = true\n"
-	    "\tbare = true\n"
-	    "[extensions]\n"
-	    "\tobjectformat = sha256\n";
-	const char *gitconfig = gitconfig_sha1;
-	char *headref_str, *path;
-	size_t i;
-
-	if (algo == GOT_HASH_SHA256)
-		gitconfig = gitconfig_sha256;
-
-	if (!got_path_dir_is_empty(repo_path))
-		return got_error(GOT_ERR_DIR_NOT_EMPTY);
-
-	for (i = 0; i < nitems(dirnames); i++) {
-		if (asprintf(&path, "%s/%s", repo_path, dirnames[i]) == -1) {
-			return got_error_from_errno("asprintf");
-		}
-		err = got_path_mkdir(path);
-		free(path);
-		if (err)
-			return err;
-	}
-
-	if (asprintf(&path, "%s/%s", repo_path, "description") == -1)
-		return got_error_from_errno("asprintf");
-	err = got_path_create_file(path, description_str);
-	free(path);
-	if (err)
-		return err;
-
-	if (asprintf(&path, "%s/%s", repo_path, GOT_HEAD_FILE) == -1)
-		return got_error_from_errno("asprintf");
-	if (asprintf(&headref_str, "%s%s", headref,
-	    head_name ? head_name : "main") == -1) {
-		free(path);
-		return got_error_from_errno("asprintf");
-	}
-	err = got_path_create_file(path, headref_str);
-	free(headref_str);
-	free(path);
-	if (err)
-		return err;
-
-	if (asprintf(&path, "%s/%s", repo_path, "config") == -1)
-		return got_error_from_errno("asprintf");
-	err = got_path_create_file(path, gitconfig);
-	free(path);
-	if (err)
-		return err;
-
-	return NULL;
-}
-
 static const struct got_error *
 match_packed_object(struct got_object_id **unique_id,
     struct got_repository *repo, const char *id_str_prefix, int obj_type)
@@ -1835,21 +1762,6 @@ retry:
 		struct got_packidx *packidx;
 		struct got_object_qid *qid;
 
-		/*
-		 * If the modification time of the 'objects/pack' directory
-		 * has changed then 'pe' could now be an invalid pointer.
-		 */
-		if (tv.tv_sec != repo->pack_path_mtime.tv_sec ||
-		    tv.tv_nsec != repo->pack_path_mtime.tv_nsec) {
-			if (++retries > max_retries) {
-				err = got_error_msg(GOT_ERR_TIMEOUT,
-				    "too many concurrent pack file "
-				    "modifications");
-				goto done;
-			}
-			goto retry;
-		}
-
 		path_packidx = pe->path;
 
 		err = got_packidx_open(&packidx, got_repo_get_fd(repo),
@@ -1881,6 +1793,24 @@ retry:
 				    &qid->id);
 				if (err)
 					goto done;
+				/*
+				 * If the modification time of the
+				 * 'objects/pack' directory has changed then
+				 * 'pe' could now be an invalid pointer.
+				 */
+				if (tv.tv_sec !=
+				    repo->pack_path_mtime.tv_sec ||
+				    tv.tv_nsec !=
+				    repo->pack_path_mtime.tv_nsec) {
+					if (++retries > max_retries) {
+						err = got_error_msg(
+						    GOT_ERR_TIMEOUT,
+						    "too many concurrent "
+						    "pack file modifications");
+						goto done;
+					}
+					goto retry;
+				}
 				if (matched_type != obj_type)
 					continue;
 			}
