@@ -43,6 +43,7 @@
 #include "got_path.h"
 
 #include "got_lib_delta.h"
+#include "got_lib_hash.h"
 #include "got_lib_object.h"
 #include "got_lib_object_idset.h"
 #include "got_lib_object_cache.h"
@@ -51,7 +52,6 @@
 #include "got_lib_repository.h"
 #include "got_lib_ratelimit.h"
 #include "got_lib_pack_create.h"
-#include "got_lib_hash.h"
 #include "got_lib_lockfile.h"
 
 #ifndef nitems
@@ -139,6 +139,81 @@ done:
 	return err;
 }
 
+static const struct got_error *
+create_temp_packfile(int *packfd, char **tmpfile_path,
+    struct got_repository *repo)
+{
+	const struct got_error *err = NULL;
+	char *path;
+
+	*packfd = -1;
+
+	if (asprintf(&path, "%s/%s/packing.pack",
+	    got_repo_get_path_git_dir(repo), GOT_OBJECTS_PACK_DIR) == -1)
+		return got_error_from_errno("asprintf");
+
+	err = got_opentemp_named_fd(tmpfile_path, packfd, path, "");
+	if (err)
+		goto done;
+
+	if (fchmod(*packfd, GOT_DEFAULT_PACK_MODE) == -1)
+		err = got_error_from_errno2("fchmod", *tmpfile_path);
+done:
+	free(path);
+	if (err) {
+		if (*packfd != -1)
+			close(*packfd);
+		*packfd = -1;
+		free(*tmpfile_path);
+		*tmpfile_path = NULL;
+	}
+	return err;
+}
+
+static const struct got_error *
+install_packfile(FILE **packfile, int *packfd, char **packfile_path,
+    char **tmpfile_path, struct got_object_id *pack_hash,
+    struct got_repository *repo)
+{
+	const struct got_error *err;
+	char *hash_str;
+
+	err = got_object_id_str(&hash_str, pack_hash);
+	if (err)
+		return err;
+
+	if (asprintf(packfile_path, "%s/%s/pack-%s.pack",
+	    got_repo_get_path_git_dir(repo), GOT_OBJECTS_PACK_DIR,
+	    hash_str) == -1) {
+		err = got_error_from_errno("asprintf");
+		goto done;
+	}
+
+	if (lseek(*packfd, 0L, SEEK_SET) == -1) {
+		err = got_error_from_errno("lseek");
+		goto done;
+	}
+
+	if (rename(*tmpfile_path, *packfile_path) == -1) {
+		err = got_error_from_errno3("rename", *tmpfile_path,
+		    *packfile_path);
+		goto done;
+	}
+
+	free(*tmpfile_path);
+	*tmpfile_path = NULL;
+
+	*packfile = fdopen(*packfd, "w");
+	if (*packfile == NULL) {
+		err = got_error_from_errno2("fdopen", *packfile_path);
+		goto done;
+	}
+	*packfd = -1;
+done:
+	free(hash_str);
+	return err;
+}
+
 const struct got_error *
 got_repo_pack_objects(FILE **packfile, struct got_object_id **pack_hash,
     struct got_reflist_head *include_refs,
@@ -150,8 +225,7 @@ got_repo_pack_objects(FILE **packfile, struct got_object_id **pack_hash,
 	const struct got_error *err = NULL;
 	struct got_object_id **ours = NULL, **theirs = NULL;
 	int nours = 0, ntheirs = 0, packfd = -1, i;
-	char *tmpfile_path = NULL, *path = NULL, *packfile_path = NULL;
-	char *sha1_str = NULL;
+	char *tmpfile_path = NULL, *packfile_path = NULL;
 	FILE *delta_cache = NULL;
 	struct got_ratelimit rl;
 
@@ -160,19 +234,9 @@ got_repo_pack_objects(FILE **packfile, struct got_object_id **pack_hash,
 
 	got_ratelimit_init(&rl, 0, 500);
 
-	if (asprintf(&path, "%s/%s/packing.pack",
-	    got_repo_get_path_git_dir(repo), GOT_OBJECTS_PACK_DIR) == -1) {
-		err = got_error_from_errno("asprintf");
-		goto done;
-	}
-	err = got_opentemp_named_fd(&tmpfile_path, &packfd, path, "");
+	err = create_temp_packfile(&packfd, &tmpfile_path, repo);
 	if (err)
-		goto done;
-
-	if (fchmod(packfd, GOT_DEFAULT_PACK_MODE) == -1) {
-		err = got_error_from_errno2("fchmod", tmpfile_path);
-		goto done;
-	}
+		return err;
 
 	delta_cache = got_opentemp();
 	if (delta_cache == NULL) {
@@ -205,42 +269,15 @@ got_repo_pack_objects(FILE **packfile, struct got_object_id **pack_hash,
 		err = got_error_from_errno("calloc");
 		goto done;
 	}
-
-	err = got_pack_create((*pack_hash)->sha1, packfd, delta_cache,
+	err = got_pack_create(*pack_hash, packfd, delta_cache,
 	    theirs, ntheirs, ours, nours, repo, loose_obj_only,
 	    0, force_refdelta, progress_cb, progress_arg, &rl,
 	    cancel_cb, cancel_arg);
 	if (err)
 		goto done;
 
-	err = got_object_id_str(&sha1_str, *pack_hash);
-	if (err)
-		goto done;
-	if (asprintf(&packfile_path, "%s/%s/pack-%s.pack",
-	    got_repo_get_path_git_dir(repo), GOT_OBJECTS_PACK_DIR,
-	    sha1_str) == -1) {
-		err = got_error_from_errno("asprintf");
-		goto done;
-	}
-
-	if (lseek(packfd, 0L, SEEK_SET) == -1) {
-		err = got_error_from_errno("lseek");
-		goto done;
-	}
-	if (rename(tmpfile_path, packfile_path) == -1) {
-		err = got_error_from_errno3("rename", tmpfile_path,
-		    packfile_path);
-		goto done;
-	}
-	free(tmpfile_path);
-	tmpfile_path = NULL;
-
-	*packfile = fdopen(packfd, "w");
-	if (*packfile == NULL) {
-		err = got_error_from_errno2("fdopen", tmpfile_path);
-		goto done;
-	}
-	packfd = -1;
+	err = install_packfile(packfile, &packfd, &packfile_path,
+	    &tmpfile_path, *pack_hash, repo);
 done:
 	for (i = 0; i < nours; i++)
 		free(ours[i]);
@@ -249,15 +286,14 @@ done:
 		free(theirs[i]);
 	free(theirs);
 	if (packfd != -1 && close(packfd) == -1 && err == NULL)
-		err = got_error_from_errno2("close", packfile_path);
+		err = got_error_from_errno2("close",
+		    packfile_path ? packfile_path : tmpfile_path);
 	if (delta_cache && fclose(delta_cache) == EOF && err == NULL)
 		err = got_error_from_errno("fclose");
 	if (tmpfile_path && unlink(tmpfile_path) == -1 && err == NULL)
 		err = got_error_from_errno2("unlink", tmpfile_path);
 	free(tmpfile_path);
 	free(packfile_path);
-	free(sha1_str);
-	free(path);
 	if (err) {
 		free(*pack_hash);
 		*pack_hash = NULL;
@@ -269,8 +305,8 @@ done:
 }
 
 const struct got_error *
-got_repo_index_pack(FILE *packfile, struct got_object_id *pack_hash,
-    struct got_repository *repo,
+got_repo_index_pack(char **idxpath, FILE *packfile,
+    struct got_object_id *pack_hash, struct got_repository *repo,
     got_pack_index_progress_cb progress_cb, void *progress_arg,
     got_cancel_cb cancel_cb, void *cancel_arg)
 {
@@ -280,14 +316,18 @@ got_repo_index_pack(FILE *packfile, struct got_object_id *pack_hash,
 	int npackfd = -1, idxfd = -1, nidxfd = -1;
 	int tmpfds[3];
 	int idxstatus, done = 0;
+	int nobj_total = 0, nobj_indexed = 0, nobj_loose = 0, nobj_resolved = 0;
 	const struct got_error *err;
 	struct imsgbuf idxibuf;
 	pid_t idxpid;
 	char *tmpidxpath = NULL;
-	char *packfile_path = NULL, *idxpath = NULL, *id_str = NULL;
+	char *packfile_path = NULL, *id_str = NULL;
 	const char *repo_path = got_repo_get_path_git_dir(repo);
 	struct stat sb;
 
+	*idxpath = NULL;
+	memset(&idxibuf, 0, sizeof(idxibuf));
+
 	for (i = 0; i < nitems(tmpfds); i++)
 		tmpfds[i] = -1;
 
@@ -334,7 +374,7 @@ got_repo_index_pack(FILE *packfile, struct got_object_id *pack_hash,
 		goto done;
 	}
 
-	if (asprintf(&idxpath, "%s/%s/pack-%s.idx",
+	if (asprintf(idxpath, "%s/%s/pack-%s.idx",
 	    repo_path, GOT_OBJECTS_PACK_DIR, id_str) == -1) {
 		err = got_error_from_errno("asprintf");
 		goto done;
@@ -355,15 +395,18 @@ got_repo_index_pack(FILE *packfile, struct got_object_id *pack_hash,
 		err = got_error_from_errno("close");
 		goto done;
 	}
-	imsg_init(&idxibuf, imsg_idxfds[0]);
+	if (imsgbuf_init(&idxibuf, imsg_idxfds[0]) == -1) {
+		err = got_error_from_errno("imsgbuf_init");
+		goto done;
+	}
+	imsgbuf_allow_fdpass(&idxibuf);
 
 	npackfd = dup(fileno(packfile));
 	if (npackfd == -1) {
 		err = got_error_from_errno("dup");
 		goto done;
 	}
-	err = got_privsep_send_index_pack_req(&idxibuf, pack_hash->sha1,
-	    npackfd);
+	err = got_privsep_send_index_pack_req(&idxibuf, pack_hash, npackfd);
 	if (err != NULL)
 		goto done;
 	npackfd = -1;
@@ -379,8 +422,6 @@ got_repo_index_pack(FILE *packfile, struct got_object_id *pack_hash,
 	}
 	done = 0;
 	while (!done) {
-		int nobj_total, nobj_indexed, nobj_loose, nobj_resolved;
-
 		if (cancel_cb) {
 			err = cancel_cb(cancel_arg);
 			if (err)
@@ -395,11 +436,18 @@ got_repo_index_pack(FILE *packfile, struct got_object_id *pack_hash,
 		if (nobj_indexed != 0) {
 			err = progress_cb(progress_arg, sb.st_size,
 			    nobj_total, nobj_indexed, nobj_loose,
-			    nobj_resolved);
+			    nobj_resolved, 0);
 			if (err)
 				break;
 		}
 	}
+	if (done) {
+		err = progress_cb(progress_arg, sb.st_size,
+		    nobj_total, nobj_indexed, nobj_loose,
+		    nobj_resolved, done);
+		if (err)
+			goto done;
+	}
 	if (close(imsg_idxfds[0]) == -1) {
 		err = got_error_from_errno("close");
 		goto done;
@@ -409,14 +457,16 @@ got_repo_index_pack(FILE *packfile, struct got_object_id *pack_hash,
 		goto done;
 	}
 
-	if (rename(tmpidxpath, idxpath) == -1) {
-		err = got_error_from_errno3("rename", tmpidxpath, idxpath);
+	if (rename(tmpidxpath, *idxpath) == -1) {
+		err = got_error_from_errno3("rename", tmpidxpath, *idxpath);
 		goto done;
 	}
 	free(tmpidxpath);
 	tmpidxpath = NULL;
 
 done:
+	if (idxibuf.w)
+		imsgbuf_clear(&idxibuf);
 	if (tmpidxpath && unlink(tmpidxpath) == -1 && err == NULL)
 		err = got_error_from_errno2("unlink", tmpidxpath);
 	if (npackfd != -1 && close(npackfd) == -1 && err == NULL)
@@ -428,7 +478,6 @@ done:
 			err = got_error_from_errno("close");
 	}
 	free(tmpidxpath);
-	free(idxpath);
 	free(packfile_path);
 	return err;
 }
@@ -529,6 +578,7 @@ got_repo_list_pack(FILE *packfile, struct got_object_id *pack_hash,
 	struct got_packidx *packidx = NULL;
 	struct got_pack *pack = NULL;
 	uint32_t nobj, i;
+	size_t digest_len = got_hash_digest_length(repo->algo);
 
 	err = got_object_id_str(&id_str, pack_hash);
 	if (err)
@@ -545,7 +595,8 @@ got_repo_list_pack(FILE *packfile, struct got_object_id *pack_hash,
 		goto done;
 	}
 
-	err = got_packidx_open(&packidx, got_repo_get_fd(repo), idxpath, 1);
+	err = got_packidx_open(&packidx, got_repo_get_fd(repo), idxpath, 1,
+	    repo->algo);
 	if (err)
 		goto done;
 
@@ -555,7 +606,7 @@ got_repo_list_pack(FILE *packfile, struct got_object_id *pack_hash,
 
 	nobj = be32toh(packidx->hdr.fanout_table[0xff]);
 	for (i = 0; i < nobj; i++) {
-		struct got_packidx_object_id *oid;
+		uint8_t *oid;
 		struct got_object_id id, base_id;
 		off_t offset, base_offset = 0;
 		uint8_t type;
@@ -567,8 +618,9 @@ got_repo_list_pack(FILE *packfile, struct got_object_id *pack_hash,
 			if (err)
 				break;
 		}
-		oid = &packidx->hdr.sorted_ids[i];
-		memcpy(id.sha1, oid->sha1, SHA1_DIGEST_LENGTH);
+		oid = packidx->hdr.sorted_ids + i * digest_len;
+		id.algo = repo->algo;
+		memcpy(id.hash, oid, digest_len);
 
 		offset = got_packidx_get_object_offset(packidx, i);
 		if (offset == -1) {
@@ -935,11 +987,19 @@ load_commit_or_tag(int *ncommits, struct got_object_idset *traversed_ids,
 			    &qid->id);
 			if (err)
 				goto done;
+			tree_id = got_object_commit_get_tree_id(commit);
 			break;
 		case GOT_OBJ_TYPE_TAG:
 			err = got_object_open_as_tag(&tag, repo, &qid->id);
 			if (err)
 				goto done;
+			/* tree_id will be set below */
+			break;
+		case GOT_OBJ_TYPE_TREE:
+			tree_id = &qid->id;
+			break;
+		case GOT_OBJ_TYPE_BLOB:
+			tree_id = NULL;
 			break;
 		default:
 			/* should not happen */
@@ -947,37 +1007,60 @@ load_commit_or_tag(int *ncommits, struct got_object_idset *traversed_ids,
 			goto done;
 		}
 
-		/* Find a tree object to scan. */
-		if (commit) {
-			tree_id = got_object_commit_get_tree_id(commit);
-		} else if (tag) {
+		if (tag) {
+			struct got_object_id *id;
+
 			obj_type = got_object_tag_get_object_type(tag);
+			while (obj_type == GOT_OBJ_TYPE_TAG) {
+				struct got_tag_object *next_tag;
+
+				id = got_object_tag_get_object_id(tag);
+				if (!got_object_idset_contains(traversed_ids,
+				    id)) {
+					err = got_object_idset_add(
+					    traversed_ids, id, NULL);
+					if (err)
+						goto done;
+				}
+
+				err = got_object_open_as_tag(&next_tag, repo,
+				    id);
+				if (err)
+					goto done;
+
+				got_object_tag_close(tag);
+				tag = next_tag;
+				obj_type = got_object_tag_get_object_type(tag);
+			}
+			id = got_object_tag_get_object_id(tag);
 			switch (obj_type) {
 			case GOT_OBJ_TYPE_COMMIT:
 				err = got_object_open_as_commit(&commit, repo,
-				    got_object_tag_get_object_id(tag));
+				    id);
 				if (err)
 					goto done;
 				tree_id = got_object_commit_get_tree_id(commit);
 				break;
 			case GOT_OBJ_TYPE_TREE:
-				tree_id = got_object_tag_get_object_id(tag);
+				tree_id = id;
 				break;
-			default:
-				/*
-				 * Tag points at something other than a
-				 * commit or tree. Leave this weird tag object
-				 * and the object it points to.
-				 */
+			case GOT_OBJ_TYPE_BLOB:
 				if (got_object_idset_contains(traversed_ids,
-				    got_object_tag_get_object_id(tag)))
+				    id))
 					break;
-				err = got_object_idset_add(traversed_ids,
-				    got_object_tag_get_object_id(tag), NULL);
+				err = got_object_idset_add(traversed_ids, id,
+				    NULL);
 				if (err)
 					goto done;
 				break;
+			default:
+				/* should not happen */
+				err = got_error(GOT_ERR_OBJ_TYPE);
+				goto done;
 			}
+		} else if (tree_id == NULL) {
+			/* Blob which has already been marked as traversed. */
+			continue;
 		}
 
 		if (tree_id) {
@@ -985,6 +1068,7 @@ load_commit_or_tag(int *ncommits, struct got_object_idset *traversed_ids,
 			    repo, cancel_cb, cancel_arg);
 			if (err)
 				break;
+			tree_id = NULL;
 		}
 
 		if (commit || tag)
@@ -1266,9 +1350,10 @@ pack_is_redundant(int *redundant, struct got_repository *repo,
 {
 	const struct got_error *err;
 	struct got_packidx *packidx;
-	struct got_packidx_object_id *pid;
+	uint8_t *pid;
 	struct got_object_id id;
 	size_t i, nobjects;
+	size_t digest_len = got_hash_digest_length(repo->algo);
 
 	*redundant = 1;
 
@@ -1278,10 +1363,11 @@ pack_is_redundant(int *redundant, struct got_repository *repo,
 
 	nobjects = be32toh(packidx->hdr.fanout_table[0xff]);
 	for (i = 0; i < nobjects; ++i) {
-		pid = &packidx->hdr.sorted_ids[i];
+		pid = packidx->hdr.sorted_ids + i * digest_len;
 
 		memset(&id, 0, sizeof(id));
-		memcpy(&id.sha1, pid->sha1, sizeof(id.sha1));
+		memcpy(&id.hash, pid, digest_len);
+		id.algo = repo->algo;
 
 		if (got_object_idset_contains(idset, &id))
 			continue;
@@ -1335,7 +1421,7 @@ repo_purge_redundant_packfiles(struct got_repository *repo,
 	int remove, redundant_packs = 0;
 
 	npacks = 0;
-	TAILQ_FOREACH(pe, &repo->packidx_paths, entry)
+	RB_FOREACH(pe, got_pathlist_head, &repo->packidx_paths)
 		npacks++;
 
 	if (npacks == 0)
@@ -1346,7 +1432,7 @@ repo_purge_redundant_packfiles(struct got_repository *repo,
 		return got_error_from_errno("calloc");
 
 	i = 0;
-	TAILQ_FOREACH(pe, &repo->packidx_paths, entry) {
+	RB_FOREACH(pe, got_pathlist_head, &repo->packidx_paths) {
 		err = got_repo_get_packidx(&packidx, pe->path, repo);
 		if (err)
 			goto done;
@@ -1406,6 +1492,8 @@ got_repo_cleanup(struct got_repository *repo,
     off_t *pack_before, off_t *pack_after,
     int *ncommits, int *nloose, int *npacked, int dry_run, int ignore_mtime,
     got_cleanup_progress_cb progress_cb, void *progress_arg,
+    got_pack_progress_cb pack_progress_cb, void *pack_progress_arg,
+    got_pack_index_progress_cb index_progress_cb, void *index_progress_arg,
     got_cancel_cb cancel_cb, void *cancel_arg)
 {
 	const struct got_error *unlock_err, *err = NULL;
@@ -1416,11 +1504,15 @@ got_repo_cleanup(struct got_repository *repo,
 	struct got_reflist_entry *re;
 	struct got_object_id **referenced_ids;
 	int i, nreferenced;
-	int npurged = 0;
+	int npurged = 0, packfd = -1;
+	char *tmpfile_path = NULL, *packfile_path = NULL, *idxpath = NULL;
+	FILE *delta_cache = NULL, *packfile = NULL;
+	struct got_object_id pack_hash;
 	time_t max_mtime = 0;
 
 	TAILQ_INIT(&refs);
 	got_ratelimit_init(&rl, 0, 500);
+	memset(&pack_hash, 0, sizeof(pack_hash));
 
 	*loose_before = 0;
 	*loose_after = 0;
@@ -1434,6 +1526,16 @@ got_repo_cleanup(struct got_repository *repo,
 	if (err)
 		return err;
 
+	err = create_temp_packfile(&packfd, &tmpfile_path, repo);
+	if (err)
+		goto done;
+
+	delta_cache = got_opentemp();
+	if (delta_cache == NULL) {
+		err = got_error_from_errno("got_opentemp");
+		goto done;
+	}
+
 	traversed_ids = got_object_idset_alloc();
 	if (traversed_ids == NULL) {
 		err = got_error_from_errno("got_object_idset_alloc");
@@ -1458,8 +1560,7 @@ got_repo_cleanup(struct got_repository *repo,
 	}
 
 	err = get_reflist_object_ids(&referenced_ids, &nreferenced,
-	    (1 << GOT_OBJ_TYPE_COMMIT) | (1 << GOT_OBJ_TYPE_TAG),
-	    &refs, repo, cancel_cb, cancel_arg);
+	    GOT_OBJ_TYPE_ANY, &refs, repo, cancel_cb, cancel_arg);
 	if (err)
 		goto done;
 
@@ -1472,6 +1573,28 @@ got_repo_cleanup(struct got_repository *repo,
 			goto done;
 	}
 
+	err = got_pack_create(&pack_hash, packfd, delta_cache,
+	    NULL, 0, referenced_ids, nreferenced, repo, 0,
+	    0, 0, pack_progress_cb, pack_progress_arg,
+	    &rl, cancel_cb, cancel_arg);
+	if (err)
+		goto done;
+
+	err = install_packfile(&packfile, &packfd, &packfile_path,
+	    &tmpfile_path, &pack_hash, repo);
+	if (err)
+		goto done;
+
+	err = got_repo_index_pack(&idxpath, packfile, &pack_hash, repo,
+	    index_progress_cb, index_progress_arg,
+	    cancel_cb, cancel_arg);
+	if (err)
+		goto done;
+
+	err = got_repo_list_packidx(&repo->packidx_paths, repo);
+	if (err)
+		goto done;
+
 	err = repo_purge_unreferenced_loose_objects(repo, traversed_ids,
 	    loose_before, loose_after, *ncommits, nloose, npacked, &npurged,
 	    dry_run, ignore_mtime, max_mtime, &rl, progress_cb, progress_arg,
@@ -1486,6 +1609,12 @@ got_repo_cleanup(struct got_repository *repo,
 	if (err)
 		goto done;
 
+	if (dry_run) {
+		if (idxpath && unlink(idxpath) == -1)
+			err = got_error_from_errno2("unlink", idxpath);
+		if (packfile_path && unlink(packfile_path) == -1 && err == NULL)
+			err = got_error_from_errno2("unlink", packfile_path);
+	}
  done:
 	if (lk) {
 		unlock_err = got_lockfile_unlock(lk, got_repo_get_fd(repo));
@@ -1494,6 +1623,16 @@ got_repo_cleanup(struct got_repository *repo,
 	}
 	if (traversed_ids)
 		got_object_idset_free(traversed_ids);
+	if (packfd != -1 && close(packfd) == -1 && err == NULL)
+		err = got_error_from_errno2("close",
+		    packfile_path ? packfile_path : tmpfile_path);
+	if (delta_cache && fclose(delta_cache) == EOF && err == NULL)
+		err = got_error_from_errno("fclose");
+	if (tmpfile_path && unlink(tmpfile_path) == -1 && err == NULL)
+		err = got_error_from_errno2("unlink", tmpfile_path);
+	free(tmpfile_path);
+	free(packfile_path);
+	free(idxpath);
 	return err;
 }
 
@@ -1522,6 +1661,7 @@ got_repo_remove_lonely_packidx(struct got_repository *repo, int dry_run,
 	packdir = fdopendir(packdir_fd);
 	if (packdir == NULL) {
 		err = got_error_from_errno("fdopendir");
+		close(packdir_fd);
 		goto done;
 	}
 
@@ -1533,7 +1673,8 @@ got_repo_remove_lonely_packidx(struct got_repository *repo, int dry_run,
 		}
 
 		if (!got_repo_is_packidx_filename(dent->d_name,
-		    strlen(dent->d_name)))
+		    strlen(dent->d_name),
+		    got_repo_get_object_format(repo)))
 			continue;
 
 		err = got_packidx_get_packfile_path(&pack_relpath,
