@@ -394,7 +394,7 @@ test_checkout_read_only() {
 	local testroot=`test_init checkout_read_only`
 	local commit_id=`git_show_head $testroot/repo`
 
-	# Make the repostiory read-only
+	# Make the repository read-only
 	chmod -R a-w $testroot/repo
 
 	echo "A  $testroot/wt/alpha" > $testroot/stdout.expected
@@ -1089,7 +1089,7 @@ run_test test_checkout_read_only
 run_test test_checkout_into_nonempty_dir
 run_test test_checkout_symlink
 run_test test_checkout_symlink_relative_wtpath
-run_test test_checkout_repo_with_unknown_extension
+run_test test_checkout_repo_with_unknown_extension	no-sha256 # XXX git fsck fails?
 run_test test_checkout_quiet
 run_test test_checkout_umask
 run_test test_checkout_ulimit_n
