@@ -28,7 +28,7 @@ test_diff_contiguous_commits()
 	local author_time=`git_show_author_time $testroot/repo`
 	local date=`date -u -r $author_time +"%a %b %e %X %Y UTC"`
 	local head_id=`git_show_head $testroot/repo`
-	local head_id_truncated=`trim_obj_id 13 $head_id`
+	local head_id_truncated=`trim_obj_id 27 $head_id`
 	local alpha_id=`get_blob_id $testroot/repo "" alpha`
 
 	cat <<EOF >$TOG_TEST_SCRIPT
@@ -36,7 +36,7 @@ SCREENDUMP
 EOF
 
 	cat <<EOF >$testroot/view.expected
-[1/20] diff $commit_id1 $head_id_truncated
+$(trim 80 "[1/20] diff $commit_id1 $head_id_truncated")
 commit $head_id (master)
 from: Flan Hacker <flan_hacker@openbsd.org>
 date: $date
@@ -76,7 +76,7 @@ EOF
 
 test_diff_arbitrary_commits()
 {
-	test_init diff_arbitrary_commits 80 18
+	test_init diff_arbitrary_commits
 
 	local commit_id1=`git_show_head $testroot/repo`
 	local alpha_id_old=`get_blob_id $testroot/repo "" alpha`
@@ -90,7 +90,7 @@ test_diff_arbitrary_commits()
 	(cd $testroot/repo && git add new)
 	git_commit $testroot/repo -m "new file"
 	local head_id=`git_show_head $testroot/repo`
-	local head_id_truncated=`trim_obj_id 13 $head_id`
+	local head_id_truncated=`trim_obj_id 27 $head_id`
 	local alpha_id=`get_blob_id $testroot/repo "" alpha`
 	local new_id=`get_blob_id $testroot/repo "" new`
 
@@ -99,7 +99,12 @@ SCREENDUMP
 EOF
 
 	cat <<EOF >$testroot/view.expected
-[1/16] diff $commit_id1 $head_id_truncated
+$(trim 80 "[1/21] diff $commit_id1 $head_id_truncated")
+M  alpha  |  1+  1-
+A  new    |  1+  0-
+
+2 files changed, 2 insertions(+), 1 deletion(-)
+
 commit - $commit_id1
 commit + $head_id
 blob - $alpha_id_old
@@ -110,12 +115,13 @@ blob + $alpha_id
 -alpha
 +modified alpha again
 blob - /dev/null
-blob + $new_id (mode 644)
+$(trim 80 "blob + $new_id (mode 644)")
 --- /dev/null
 +++ new
 @@ -0,0 +1 @@
 +new file
 
+
 (END)
 EOF
 
@@ -180,7 +186,7 @@ SCREENDUMP
 EOF
 
 	cat <<EOF >$testroot/view.expected
-[1/20] diff $id6 $id7
+$(trim 94 "[1/20] diff $id6 $id7")
 commit $id7
 from: Flan Hacker <flan_hacker@openbsd.org>
 date: $date7
@@ -228,7 +234,7 @@ SCREENDUMP
 EOF
 
 	cat <<EOF >$testroot/view.expected
-[1/20] diff $id25 $id26
+$(trim 94 "[1/20] diff $id25 $id26")
 commit $id26
 from: Flan Hacker <flan_hacker@openbsd.org>
 date: $date26
@@ -276,7 +282,7 @@ SCREENDUMP
 EOF
 
 	cat <<EOF >$testroot/view.expected
-[1/20] diff $id6 $id7
+$(trim 94 "[1/20] diff $id6 $id7")
 commit $id7
 from: Flan Hacker <flan_hacker@openbsd.org>
 date: $date7
@@ -364,7 +370,7 @@ test_diff_commit_keywords()
 	local date=$(date -u -r $(pop_idx 2 $@) +"%a %b %e %X %Y UTC")
 
 	cat <<-EOF >$testroot/view.expected
-	[1/20] diff $lhs_id $rhs_id
+	$(trim 120 "[1/20] diff $lhs_id $rhs_id")
 	commit $rhs_id
 	from: Flan Hacker <flan_hacker@openbsd.org>
 	date: $date
@@ -404,7 +410,11 @@ test_diff_commit_keywords()
 	rhs_id=$(pop_idx 8 $ids)
 
 	cat <<-EOF >$testroot/view.expected
-	[1/10] diff $lhs_id $rhs_id
+	$(trim 120 "[1/14] diff $lhs_id $rhs_id")
+	M  alpha  |  1+  1-
+
+	1 file changed, 1 insertion(+), 1 deletion(-)
+
 	commit - $lhs_id
 	commit + $rhs_id
 	blob - $(pop_idx 5 $alpha_ids)
@@ -423,10 +433,6 @@ test_diff_commit_keywords()
 
 
 
-
-
-
-
 	(END)
 	EOF
 
@@ -499,7 +505,7 @@ test_diff_horizontal_scroll()
 	local author_time=`git_show_author_time $testroot/repo`
 	local date=`date -u -r $author_time +"%a %b %e %X %Y UTC"`
 	local head_id=`git_show_head $testroot/repo`
-	local head_id_truncated=`trim_obj_id 13 $head_id`
+	local head_id_truncated=`trim_obj_id 27 $head_id`
 	local alpha_id=`get_blob_id $testroot/repo "" alpha`
 
 	cat <<EOF >$TOG_TEST_SCRIPT
@@ -508,7 +514,7 @@ SCREENDUMP
 EOF
 
 	cat <<EOF >$testroot/view.expected
-[1/20] diff $commit_id1 $head_id_truncated
+$(trim 80 "[1/20] diff $commit_id1 $head_id_truncated")
  $head_id (master)
 Flan Hacker <flan_hacker@openbsd.org>
 $date
@@ -546,9 +552,234 @@ EOF
 	test_done "$testroot" "$ret"
 }
 
+test_diff_p_keymap()
+{
+	test_init diff_p_keymap
+
+	local id_root=$(git_show_head $testroot/repo)
+	local alpha_root=$(get_blob_id $testroot/repo "" alpha)
+	local beta_root=$(get_blob_id $testroot/repo "" beta)
+
+	echo "modified alpha" > $testroot/repo/alpha
+	git_commit $testroot/repo -m "modified alpha"
+	local id_mid=$(git_show_head $testroot/repo)
+	local alpha_mid=$(get_blob_id $testroot/repo "" alpha)
+	local beta_mid=$(get_blob_id $testroot/repo "" beta)
+
+	echo "modified beta" > $testroot/repo/beta
+	git_commit $testroot/repo -m "modified beta"
+	local id_head=$(git_show_head $testroot/repo)
+	local beta_head=$(get_blob_id $testroot/repo "" beta)
+	local author_time=$(git_show_author_time $testroot/repo)
+	local date_head=$(date -u -r $author_time +"%a %b %e %X %Y UTC")
+
+	cat <<-EOF >$TOG_TEST_SCRIPT
+	p
+	SCREENDUMP
+	EOF
+
+	cat <<EOF >$testroot/content.expected
+commit $id_head (master)
+from: Flan Hacker <flan_hacker@openbsd.org>
+date: $date_head
+
+modified beta
+
+M  beta  |  1+  1-
+
+1 file changed, 1 insertion(+), 1 deletion(-)
+
+commit - $id_mid
+commit + $id_head
+blob - $beta_mid
+blob + $beta_head
+--- beta
++++ beta
+@@ -1 +1 @@
+-beta
++modified beta
+EOF
+
+	# test diff of parent:child commit has commit info in patch file
+	cd $testroot/repo && tog diff :head:- :head
+	local patchpath=$(tail -1 $testroot/view | cut -d' ' -f 5)
+
+	if [ ! -e "$patchpath" ]; then
+		echo "tog diff 'p' keymap failed to write patch file" >&2
+		test_done "$testroot" 1
+		return 1
+	fi
+
+	mv $patchpath $testroot/content
+	ret=$?
+	if [ $ret -ne 0 ]; then
+		echo "renaming patch file failed unexpectedly" >&2
+		test_done "$testroot" $ret
+		return 1
+	fi
+
+	cmp -s $testroot/content.expected $testroot/content
+	ret=$?
+	if [ $ret -ne 0 ]; then
+		diff -u $testroot/content.expected $testroot/content
+		test_done "$testroot" $ret
+		return 1
+	fi
+
+	cat <<EOF >$testroot/content.expected
+M  alpha  |  1+  1-
+M  beta   |  1+  1-
+
+2 files changed, 2 insertions(+), 2 deletions(-)
+
+commit - $id_root
+commit + $id_head
+blob - $alpha_root
+blob + $alpha_mid
+--- alpha
++++ alpha
+@@ -1 +1 @@
+-alpha
++modified alpha
+blob - $beta_root
+blob + $beta_head
+--- beta
++++ beta
+@@ -1 +1 @@
+-beta
++modified beta
+EOF
+
+	# test diff of arbitrary commits displays diffstat in patch file
+	cd $testroot/repo && tog diff :head:-2 :head
+	patchpath=$(tail -1 $testroot/view | cut -d' ' -f 5)
+
+	if [ ! -e "$patchpath" ]; then
+		echo "tog diff 'p' keymap failed to write patch file" >&2
+		test_done "$testroot" 1
+		return 1
+	fi
+
+	mv -f $patchpath $testroot/content
+	ret=$?
+	if [ $ret -ne 0 ]; then
+		echo "renaming patch file failed unexpectedly" >&2
+		test_done "$testroot" $ret
+		return 1
+	fi
+
+	cmp -s $testroot/content.expected $testroot/content
+	ret=$?
+	if [ $ret -ne 0 ]; then
+		diff -u $testroot/content.expected $testroot/content
+		test_done "$testroot" $ret
+		return 1
+	fi
+
+	test_done "$testroot" $ret
+}
+
+test_diff_blobs()
+{
+	test_init diff_blobs 148 14
+
+	local blob_alpha_root=$(get_blob_id $testroot/repo "" alpha)
+
+	echo "new alpha" > $testroot/repo/alpha
+	git_commit $testroot/repo -m "new alpha"
+
+	local blob_alpha_head=$(get_blob_id $testroot/repo "" alpha)
+
+	local short_alpha_root=$(printf '%.10s' $blob_alpha_root)
+	local short_alpha_head=$(printf '%.10s' $blob_alpha_head)
+
+	cat <<EOF >$TOG_TEST_SCRIPT
+SCREENDUMP
+EOF
+
+	cat <<EOF >$testroot/view.expected
+[1/12] diff $blob_alpha_root $blob_alpha_head
+M  $short_alpha_root -> $short_alpha_head  |  1+  1-
+
+1 file changed, 1 insertion(+), 1 deletion(-)
+
+blob - $blob_alpha_root
+blob + $blob_alpha_head
+--- $blob_alpha_root
++++ $blob_alpha_head
+@@ -1 +1 @@
+-alpha
++new alpha
+
+(END)
+EOF
+
+	cd $testroot/repo && tog diff $blob_alpha_root $blob_alpha_head
+	cmp -s $testroot/view.expected $testroot/view
+	ret=$?
+	if [ $ret -ne 0 ]; then
+		diff -u $testroot/view.expected $testroot/view
+		test_done "$testroot" $ret
+		return 1
+	fi
+
+	test_done "$testroot" $ret
+}
+
+test_diff_trees()
+{
+	test_init diff_trees 148 16
+
+	local tree_gamma_root=$(get_blob_id $testroot/repo "" gamma/)
+	local blob_delta_root=$(get_blob_id $testroot/repo gamma delta)
+
+	echo "new delta" > $testroot/repo/gamma/delta
+	git_commit $testroot/repo -m "new delta"
+
+	local tree_gamma_head=$(get_blob_id $testroot/repo "" gamma/)
+	local blob_delta_head=$(get_blob_id $testroot/repo gamma delta)
+
+	cat <<EOF >$TOG_TEST_SCRIPT
+SCREENDUMP
+EOF
+
+	cat <<EOF >$testroot/view.expected
+[1/14] diff $tree_gamma_root $tree_gamma_head
+M  delta  |  1+  1-
+
+1 file changed, 1 insertion(+), 1 deletion(-)
+
+tree - $tree_gamma_root
+tree + $tree_gamma_head
+blob - $blob_delta_root
+blob + $blob_delta_head
+--- delta
++++ delta
+@@ -1 +1 @@
+-delta
++new delta
+
+(END)
+EOF
+
+	cd $testroot/repo && tog diff $tree_gamma_root $tree_gamma_head
+	cmp -s $testroot/view.expected $testroot/view
+	ret=$?
+	if [ $ret -ne 0 ]; then
+		diff -u $testroot/view.expected $testroot/view
+		test_done "$testroot" $ret
+		return 1
+	fi
+
+	test_done "$testroot" $ret
+}
+
 test_parseargs "$@"
 run_test test_diff_contiguous_commits
 run_test test_diff_arbitrary_commits
 run_test test_diff_J_keymap
 run_test test_diff_commit_keywords
 run_test test_diff_horizontal_scroll
+run_test test_diff_p_keymap
+run_test test_diff_blobs
+run_test test_diff_trees
